import os

import pytest
from nbformat.v4.nbbase import new_markdown_cell, new_notebook

from jupytext import TextFileContentsManager, read, write
from jupytext.cli import jupytext


@pytest.mark.requires_sphinx_gallery
def test_rst2md(tmpdir, cwd_tmpdir):
    tmp_py = "notebook.py"
    tmp_ipynb = "notebook.ipynb"

    # Write notebook in sphinx format
    nb = new_notebook(
        cells=[
            new_markdown_cell("A short sphinx notebook"),
            new_markdown_cell(":math:`1+1`"),
        ]
    )
    write(nb, tmp_py, fmt="py:sphinx")

    jupytext(
        [
            tmp_py,
            "--from",
            "py:sphinx",
            "--to",
            "ipynb",
            "--opt",
            "rst2md=True",
            "--opt",
            "cell_metadata_filter=-all",
        ]
    )

    assert os.path.isfile(tmp_ipynb)
    nb = read(tmp_ipynb)

    assert nb.metadata["jupytext"]["cell_metadata_filter"] == "-all"
    assert nb.metadata["jupytext"]["rst2md"] is False

    # Was rst to md conversion effective?
    assert nb.cells[2].source == "$1+1$"


@pytest.mark.requires_sphinx_gallery
def test_rst2md_option(tmpdir):
    tmp_py = str(tmpdir.join("notebook.py"))

    # Write notebook in sphinx format
    nb = new_notebook(
        cells=[
            new_markdown_cell("A short sphinx notebook"),
            new_markdown_cell(":math:`1+1`"),
        ]
    )
    write(nb, tmp_py, fmt="py:sphinx")

    cm = TextFileContentsManager()
    cm.sphinx_convert_rst2md = True
    cm.root_dir = str(tmpdir)

    nb2 = cm.get("notebook.py")["content"]

    # Was rst to md conversion effective?
    assert nb2.cells[2].source == "$1+1$"
    assert nb2.metadata["jupytext"]["rst2md"] is False



import pytest

import jupytext
from jupytext.compare import compare_notebooks


def test_as_version_has_appropriate_type():
    with pytest.raises(TypeError):
        jupytext.read("script.py", "py:percent")


def test_read_file_with_explicit_fmt(tmpdir):
    tmp_py = str(tmpdir.join("notebook.py"))

    with open(tmp_py, "w") as fp:
        fp.write(
            """# %%
1 + 1
"""
        )

    nb1 = jupytext.read(tmp_py)
    nb2 = jupytext.read(tmp_py, fmt="py:percent")

    compare_notebooks(nb2, nb1)



import pytest

from jupytext.cli import jupytext


@pytest.mark.requires_myst
@pytest.mark.requires_black
def test_cli_black_myst(
    tmp_path,
    no_jupytext_version_number,
    text="""---
jupytext:
  formats: md:myst
  text_representation:
    extension: .md
    format_name: myst
    format_version: 0.13
kernelspec:
  display_name: Python 3 (ipykernel)
  language: python
  name: python3
---

# Quantum yield efficiency of photosynthesis

```{code-cell} python
# I'm some code
x = 1
```
""",
):
    tmp_md = tmp_path / "notebook.md"
    tmp_md.write_text(text)

    jupytext(["--pipe", "black", str(tmp_md)])

    assert tmp_md.read_text() == text



from pathlib import Path

import pytest

from jupytext import read


def documentation_files():
    for path in (Path(__file__).parent / "../../../docs").iterdir():
        if path.suffix == ".md":
            yield path


@pytest.mark.parametrize(
    "doc_file",
    documentation_files(),
    ids=[doc_file.stem for doc_file in documentation_files()],
)
def test_doc_files_are_notebooks(doc_file):
    nb = read(doc_file)

    # count how many cell types
    counts = {"markdown": 0, "raw": 0, "code": 0}
    for cell in nb.cells:
        counts[cell.cell_type] += 1

    assert counts["raw"] <= 1



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------

project = "Jupytext"
copyright = "2018-2023, The Jupytext Team"
author = "The Jupytext Team"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx_copybutton", "myst_parser"]

# Auto-generated header anchors
myst_heading_anchors = 3

html_context = {
    "display_github": True,  # Integrate GitHub
    "github_user": "mwouts",  # Username
    "github_repo": "jupytext",  # Repo name
    "github_version": "main",  # Version
    "conf_py_path": "/docs/",  # Path in the checkout to the docs root
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "alabaster"

html_sidebars = {"**": ["about.html", "navigation.html", "searchbox.html"]}

html_theme_options = {
    "github_button": True,
    "github_banner": True,
    "github_user": "mwouts",
    "github_repo": "jupytext",
    "github_type": "star",
    "logo": "logo.svg",
    "show_relbars": True,
}

pygments_style = "sphinx"
master_doc = "index"  # Makes `index.md` the main file

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Output file base name for HTML help builder.
htmlhelp_basename = "jupytext"



from nbformat.v4.nbbase import new_raw_cell

from jupytext import reads, writes
from jupytext.cli import jupytext


def test_set_cell_markers_cli(tmpdir, cwd_tmpdir):
    tmpdir.join("test.py").write("# %% [markdown]\n# A Markdown cell\n")
    jupytext(["--format-options", 'cell_markers="""', "test.py"])
    py = tmpdir.join("test.py").read()
    assert py.endswith('# %% [markdown]\n"""\nA Markdown cell\n"""\n')


def test_add_cell_to_script_with_cell_markers(
    no_jupytext_version_number,
    py='''# ---
# jupyter:
#   jupytext:
#     formats: py:percent
#     cell_markers: '"""'
# ---
''',
):
    nb = reads(py, fmt="py:percent")
    nb.cells = [new_raw_cell("A raw cell")]
    py2 = writes(nb, fmt="py:percent")
    assert py2.endswith(
        '''# %% [raw]
"""
A raw cell
"""
'''
    )



import os
from copy import deepcopy
from shutil import copyfile

import pytest
from nbformat.v4.nbbase import new_code_cell, new_notebook

from jupytext import read, write
from jupytext.cli import jupytext, pipe_notebook, system
from jupytext.combine import black_invariant
from jupytext.compare import compare, compare_cells, compare_notebooks
from jupytext.header import _DEFAULT_NOTEBOOK_METADATA


@pytest.mark.requires_black
def test_apply_black_on_python_notebooks(tmpdir, cwd_tmpdir, ipynb_py_file):
    if "cell metadata" in ipynb_py_file:
        pytest.skip()
    copyfile(ipynb_py_file, "notebook.ipynb")

    jupytext(args=["notebook.ipynb", "--to", "py:percent"])
    system("black", "notebook.py")
    jupytext(args=["notebook.py", "--to", "ipynb", "--update"])

    nb1 = read(ipynb_py_file)
    nb2 = read("notebook.ipynb")
    nb3 = read("notebook.py")

    assert len(nb1.cells) == len(nb2.cells)
    assert len(nb1.cells) == len(nb3.cells)
    for c1, c2 in zip(nb1.cells, nb2.cells):
        # same content (almost)
        assert black_invariant(c1.source) == black_invariant(c2.source)
        # python representation is pep8
        assert "lines_to_next_cell" not in c2.metadata
        # outputs are preserved
        assert c1.cell_type == c2.cell_type
        if c1.cell_type == "code":
            compare(c1.outputs, c2.outputs)

    compare(nb1.metadata, nb2.metadata)


def test_black_invariant():
    text_org = """long_string = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" \\
              "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
"""
    text_black = """long_string = (
    "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
    "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
)
"""
    assert black_invariant(text_org) == black_invariant(text_black)


@pytest.mark.requires_black
def test_pipe_into_black():
    nb_org = new_notebook(cells=[new_code_cell("1        +1", id="cell-id")])
    nb_dest = new_notebook(cells=[new_code_cell("1 + 1", id="cell-id")])

    nb_pipe = pipe_notebook(nb_org, "black")
    compare_notebooks(
        nb_pipe, nb_dest, allow_expected_differences=False, compare_ids=True
    )


@pytest.mark.requires_autopep8
def test_pipe_into_autopep8():
    nb_org = new_notebook(cells=[new_code_cell("1        +1", id="cell-id")])
    nb_dest = new_notebook(cells=[new_code_cell("1 + 1", id="cell-id")])

    nb_pipe = pipe_notebook(nb_org, "autopep8 -")
    compare_notebooks(
        nb_pipe, nb_dest, allow_expected_differences=False, compare_ids=True
    )


@pytest.mark.requires_flake8
def test_pipe_into_flake8():
    # Notebook OK
    nb = new_notebook(cells=[new_code_cell("# correct code\n1 + 1")])
    pipe_notebook(nb, "flake8", update=False)

    # Notebook not OK
    nb = new_notebook(cells=[new_code_cell("incorrect code")])
    with pytest.raises(SystemExit):
        pipe_notebook(nb, "flake8", update=False)


@pytest.mark.requires_black
@pytest.mark.requires_flake8
def test_apply_black_through_jupytext(tmpdir, python_notebook):
    # Load real notebook metadata to get the 'auto' extension in --pipe-fmt to work
    metadata = python_notebook.metadata

    nb_org = new_notebook(
        cells=[new_code_cell("1        +1", id="cell-id")], metadata=metadata
    )
    nb_black = new_notebook(
        cells=[new_code_cell("1 + 1", id="cell-id")], metadata=metadata
    )

    tmp_ipynb = str(tmpdir.mkdir("notebook_folder").join("notebook.ipynb"))
    tmp_py = str(tmpdir.mkdir("script_folder").join("notebook.py"))

    # Black in place
    write(nb_org, tmp_ipynb)
    jupytext([tmp_ipynb, "--pipe", "black"])
    nb_now = read(tmp_ipynb)
    compare_notebooks(nb_now, nb_black, compare_ids=True)

    # Write to another folder using dots
    write(nb_org, tmp_ipynb)
    jupytext([tmp_ipynb, "--to", "../script_folder//py:percent", "--pipe", "black"])
    assert os.path.isfile(tmp_py)
    nb_now = read(tmp_py)
    nb_now.metadata = metadata
    compare_notebooks(nb_now, nb_black)
    os.remove(tmp_py)

    # Map to another folder based on file name
    write(nb_org, tmp_ipynb)
    jupytext(
        [
            tmp_ipynb,
            "--from",
            "notebook_folder//ipynb",
            "--to",
            "script_folder//py:percent",
            "--pipe",
            "black",
            "--check",
            "flake8",
        ]
    )
    assert os.path.isfile(tmp_py)
    nb_now = read(tmp_py)
    nb_now.metadata = metadata
    compare_notebooks(nb_now, nb_black)


@pytest.mark.requires_black
def test_apply_black_and_sync_on_paired_notebook(tmpdir, cwd_tmpdir, python_notebook):
    # Load real notebook metadata to get the 'auto' extension in --pipe-fmt to work
    metadata = python_notebook.metadata
    metadata["jupytext"] = {"formats": "ipynb,py"}
    assert "language_info" in metadata

    nb_org = new_notebook(
        cells=[new_code_cell("1        +1", id="cell-id")], metadata=metadata
    )
    nb_black = new_notebook(
        cells=[new_code_cell("1 + 1", id="cell-id")], metadata=metadata
    )

    # Black in place
    write(nb_org, "notebook.ipynb")
    jupytext(["notebook.ipynb", "--pipe", "black", "--sync"])

    nb_now = read("notebook.ipynb")
    compare_notebooks(nb_now, nb_black, compare_ids=True)
    assert "language_info" in nb_now.metadata

    nb_now = read("notebook.py")
    nb_now.metadata["jupytext"].pop("text_representation")
    nb_black.metadata = {
        key: nb_black.metadata[key]
        for key in nb_black.metadata
        if key in _DEFAULT_NOTEBOOK_METADATA.split(",")
    }
    compare_notebooks(nb_now, nb_black)


@pytest.mark.requires_black
def test_apply_black_on_markdown_notebook(tmpdir):
    text = """---
jupyter:
  kernelspec:
    display_name: Python 3
    language: python
    name: python3
  language_info:
    codemirror_mode:
      name: ipython
      version: 3
    file_extension: .py
    mimetype: text/x-python
    name: python
    nbconvert_exporter: python
    pygments_lexer: ipython3
    version: 3.7.4
---

```python
1    +     \
2+3\
+4
```
"""
    tmp_md = str(tmpdir.join("test.md"))
    with open(tmp_md, "w") as fp:
        fp.write(text)

    jupytext([tmp_md, "--pipe", "black"])

    nb = read(tmp_md)
    compare_cells(nb.cells, [new_code_cell("1 + 2 + 3 + 4")], compare_ids=False)


@pytest.mark.requires_black
def test_black_through_tempfile(
    tmpdir,
    text="""```python
1 +    2 \
+ 3
```
""",
    black="""```python
1 + 2 + 3
```
""",
):
    tmp_md = str(tmpdir.join("notebook.md"))
    with open(tmp_md, "w") as fp:
        fp.write(text)
    jupytext([tmp_md, "--pipe", "black {}"])
    with open(tmp_md) as fp:
        compare(fp.read(), black)


@pytest.mark.requires_black
def test_pipe_black_removes_lines_to_next_cell_metadata(
    tmpdir,
    cwd_tmpdir,
    text="""# %%
def func():
    return 42
# %%
func()""",
):
    tmpdir.join("notebook.py").write(text)
    jupytext(["--set-formats", "ipynb,py:percent", "notebook.py"])

    nb = read(tmpdir.join("notebook.ipynb"))
    assert nb.cells[0].metadata["lines_to_next_cell"] == 0

    jupytext(["--sync", "notebook.py", "--pipe", "black"])
    nb = read(tmpdir.join("notebook.ipynb"))
    assert "lines_to_next_cell" not in nb.cells[0].metadata

    new_text = tmpdir.join("notebook.py").read()
    assert "\n\n# %%\nfunc()" in new_text


@pytest.mark.requires_black
@pytest.mark.parametrize(
    "code,black_should_fail",
    [("myvar = %dont_format_me", False), ("incomplete_instruction = (...", True)],
)
def test_pipe_black_uses_warn_only_781(
    tmpdir, cwd_tmpdir, code, black_should_fail, python_notebook, capsys
):
    nb = python_notebook
    nb.cells.append(new_code_cell(code))
    write(nb, "notebook.ipynb")

    if not black_should_fail:
        jupytext(["--pipe", "black", "notebook.ipynb"])
        return

    with pytest.raises(SystemExit):
        jupytext(["--pipe", "black", "notebook.ipynb"])

    out, err = capsys.readouterr()
    assert "Error: The command 'black -' exited with code" in err
    assert "--warn-only" in err

    # With warn-only we just get a warning
    jupytext(["--pipe", "black", "notebook.ipynb", "--warn-only"])
    out, err = capsys.readouterr()
    assert "Warning: The command 'black -' exited with code" in err

    # If black fails the notebook should be left unchanged
    actual = read("notebook.ipynb")
    compare_notebooks(actual, nb)


@pytest.mark.requires_black
def test_pipe_black_preserve_outputs(notebook_with_outputs, tmpdir, cwd_tmpdir, capsys):
    write(notebook_with_outputs, "test.ipynb")
    jupytext(["--pipe", "black", "test.ipynb"])

    # Outputs are preserved
    nb = read("test.ipynb")
    expected = deepcopy(notebook_with_outputs)
    expected.cells[0].source = "1 + 1"
    compare_notebooks(nb, expected)

    # No mention of --update
    out, err = capsys.readouterr()
    assert not err
    assert "replaced" not in out
    assert "--update" not in out



import inspect
import os
import re
import shutil
import time

import pytest
from jupyter_server.utils import ensure_async
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook
from tornado.web import HTTPError

import jupytext
from jupytext.cli import jupytext as jupytext_cli
from jupytext.compare import compare, compare_notebooks, notebook_model
from jupytext.formats import auto_ext_from_metadata, read_format_from_metadata
from jupytext.header import header_to_metadata_and_cell
from jupytext.jupytext import read, write, writes
from jupytext.kernels import kernelspec_from_language

pytestmark = pytest.mark.asyncio


async def test_rename(tmpdir, cm):
    org_file = str(tmpdir.join("notebook.ipynb"))
    new_file = str(tmpdir.join("new.ipynb"))
    jupytext.write(new_notebook(), org_file)

    cm.root_dir = str(tmpdir)
    await ensure_async(cm.rename_file("notebook.ipynb", "new.ipynb"))

    assert os.path.isfile(new_file)
    assert not os.path.isfile(org_file)


async def test_rename_inconsistent_path(tmpdir, cm):
    org_file = str(tmpdir.join("notebook_suffix.ipynb"))
    new_file = str(tmpdir.join("new.ipynb"))
    jupytext.write(
        new_notebook(metadata={"jupytext": {"formats": "_suffix.ipynb"}}), org_file
    )

    cm.root_dir = str(tmpdir)
    # Read notebook, and learn about its format
    await ensure_async(cm.get("notebook_suffix.ipynb"))
    with pytest.raises(HTTPError):
        await ensure_async(cm.rename_file("notebook_suffix.ipynb", "new.ipynb"))

    assert not os.path.isfile(new_file)
    assert os.path.isfile(org_file)


async def test_pair_unpair_notebook(tmpdir, cm):
    tmp_ipynb = "notebook.ipynb"
    tmp_md = "notebook.md"

    nb = new_notebook(
        metadata={
            "kernelspec": {
                "display_name": "Python3",
                "language": "python",
                "name": "python3",
            }
        },
        cells=[
            new_code_cell(
                "1 + 1",
                outputs=[
                    {
                        "data": {"text/plain": ["2"]},
                        "execution_count": 1,
                        "metadata": {},
                        "output_type": "execute_result",
                    }
                ],
            )
        ],
    )

    cm.root_dir = str(tmpdir)

    # save notebook
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))
    assert not os.path.isfile(str(tmpdir.join(tmp_md)))

    # pair notebook
    nb["metadata"]["jupytext"] = {"formats": "ipynb,md"}
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))
    assert os.path.isfile(str(tmpdir.join(tmp_md)))

    # reload and get outputs
    nb2 = (await ensure_async(cm.get(tmp_md)))["content"]
    compare_notebooks(nb, nb2)

    # unpair and save as md
    del nb["metadata"]["jupytext"]
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_md))
    nb2 = (await ensure_async(cm.get(tmp_md)))["content"]

    # we get no outputs here
    compare_notebooks(nb, nb2, compare_outputs=False)
    assert len(nb2.cells[0]["outputs"]) == 0


async def test_load_save_rename(ipynb_py_R_jl_file, cm, tmpdir):
    tmp_ipynb = "notebook.ipynb"
    tmp_rmd = "notebook.Rmd"

    cm.formats = "ipynb,Rmd"
    cm.root_dir = str(tmpdir)
    cm.delete_to_trash = False

    # open ipynb, save Rmd, reopen
    nb = jupytext.read(ipynb_py_R_jl_file)
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_rmd))
    nb_rmd = await ensure_async(cm.get(tmp_rmd))
    compare_notebooks(nb_rmd["content"], nb, "Rmd")

    # save ipynb
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # rename_file ipynb
    await ensure_async(cm.rename_file(tmp_ipynb, "new.ipynb"))
    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_rmd)))

    assert os.path.isfile(str(tmpdir.join("new.ipynb")))
    assert os.path.isfile(str(tmpdir.join("new.Rmd")))

    # delete one file, test that we can still read and rename_file it
    await ensure_async(cm.delete("new.Rmd"))
    assert not os.path.isfile(str(tmpdir.join("new.Rmd")))
    model = await ensure_async(cm.get("new.ipynb", content=False))
    assert "last_modified" in model
    await ensure_async(cm.save(model=notebook_model(nb), path="new.ipynb"))
    assert os.path.isfile(str(tmpdir.join("new.Rmd")))

    await ensure_async(cm.delete("new.Rmd"))
    await ensure_async(cm.rename_file("new.ipynb", tmp_ipynb))

    assert os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_rmd)))
    assert not os.path.isfile(str(tmpdir.join("new.ipynb")))
    assert not os.path.isfile(str(tmpdir.join("new.Rmd")))


async def test_save_load_paired_md_notebook(ipynb_py_R_jl_file, cm, tmpdir):
    tmp_ipynb = "notebook.ipynb"
    tmp_md = "notebook.md"

    cm.root_dir = str(tmpdir)

    # open ipynb, save with cm, reopen
    nb = jupytext.read(ipynb_py_R_jl_file)
    nb.metadata["jupytext"] = {"formats": "ipynb,md"}

    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))
    nb_md = await ensure_async(cm.get(tmp_md))

    compare_notebooks(nb_md["content"], nb, "md")
    assert nb_md["content"].metadata["jupytext"]["formats"] == "ipynb,md"


async def test_pair_plain_script(percent_file, cm, tmpdir, caplog):
    tmp_py = "notebook.py"
    tmp_ipynb = "notebook.ipynb"

    cm.root_dir = str(tmpdir)

    # open py file, pair, save with cm
    nb = jupytext.read(percent_file)
    nb.metadata["jupytext"]["formats"] = "ipynb,py:hydrogen"
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_py))

    # assert "'Include Metadata' is off" in caplog.text

    assert os.path.isfile(str(tmpdir.join(tmp_py)))
    assert os.path.isfile(str(tmpdir.join(tmp_ipynb)))

    # Make sure we've not changed the script
    with open(percent_file) as fp:
        script = fp.read()

    with open(str(tmpdir.join(tmp_py))) as fp:
        script2 = fp.read()

    compare(script2, script)

    # reopen py file with the cm
    nb2 = (await ensure_async(cm.get(tmp_py)))["content"]
    compare_notebooks(nb2, nb)
    assert nb2.metadata["jupytext"]["formats"] == "ipynb,py:hydrogen"

    # remove the pairing and save
    del nb.metadata["jupytext"]["formats"]
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_py))

    # reopen py file with the cm
    nb2 = (await ensure_async(cm.get(tmp_py)))["content"]
    compare_notebooks(nb2, nb)
    assert "formats" not in nb2.metadata["jupytext"]


async def test_load_save_rename_nbpy(ipynb_py_file, cm, tmpdir):
    tmp_ipynb = "notebook.ipynb"
    tmp_nbpy = "notebook.nb.py"

    cm.formats = "ipynb,.nb.py"
    cm.root_dir = str(tmpdir)

    # open ipynb, save nb.py, reopen
    nb = jupytext.read(ipynb_py_file)
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_nbpy))
    nbpy = await ensure_async(cm.get(tmp_nbpy))
    compare_notebooks(nbpy["content"], nb)

    # save ipynb
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # rename_file nbpy
    await ensure_async(cm.rename_file(tmp_nbpy, "new.nb.py"))
    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_nbpy)))

    assert os.path.isfile(str(tmpdir.join("new.ipynb")))
    assert os.path.isfile(str(tmpdir.join("new.nb.py")))

    # rename_file to a non-matching pattern
    with pytest.raises(HTTPError):
        await ensure_async(cm.rename_file(tmp_nbpy, "suffix_missing.py"))


async def test_load_save_py_freeze_metadata(python_file, cm, tmpdir):
    if "light" in python_file:
        pytest.skip()

    tmp_nbpy = "notebook.py"

    cm.root_dir = str(tmpdir)

    # read original file
    with open(python_file) as fp:
        text_py = fp.read()

    # write to tmp_nbpy
    with open(str(tmpdir.join(tmp_nbpy)), "w") as fp:
        fp.write(text_py)

    # open and save notebook
    nb = (await ensure_async(cm.get(tmp_nbpy)))["content"]
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_nbpy))

    with open(str(tmpdir.join(tmp_nbpy))) as fp:
        text_py2 = fp.read()

    compare(text_py2, text_py)


async def test_load_text_notebook(tmpdir, cm):
    cm.root_dir = str(tmpdir)

    nbpy = "text.py"
    with open(str(tmpdir.join(nbpy)), "w") as fp:
        fp.write("# %%\n1 + 1\n")

    py_model = await ensure_async(cm.get(nbpy, content=False))
    assert py_model["type"] == "notebook"
    assert py_model["content"] is None

    py_model = await ensure_async(cm.get(nbpy, content=True))
    assert py_model["type"] == "notebook"
    assert "cells" in py_model["content"]

    # The model returned by the CM should match that of a classical ipynb notebook
    nb_model = dict(
        type="notebook", content=new_notebook(cells=[new_markdown_cell("A cell")])
    )
    await ensure_async(cm.save(nb_model, "notebook.ipynb"))
    nb_model = await ensure_async(cm.get("notebook.ipynb", content=True))
    for key in ["format", "mimetype", "type"]:
        assert nb_model[key] == py_model[key], key


async def test_load_save_rename_notebook_with_dot(ipynb_py_file, cm, tmpdir):
    tmp_ipynb = "1.notebook.ipynb"
    tmp_nbpy = "1.notebook.py"

    cm.formats = "ipynb,py"
    cm.root_dir = str(tmpdir)

    # open ipynb, save nb.py, reopen
    nb = jupytext.read(ipynb_py_file)
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_nbpy))
    nbpy = await ensure_async(cm.get(tmp_nbpy))
    compare_notebooks(nbpy["content"], nb)

    # save ipynb
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # rename_file py
    await ensure_async(cm.rename_file(tmp_nbpy, "2.new_notebook.py"))
    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_nbpy)))

    assert os.path.isfile(str(tmpdir.join("2.new_notebook.ipynb")))
    assert os.path.isfile(str(tmpdir.join("2.new_notebook.py")))


async def test_load_save_rename_nbpy_default_config(ipynb_py_file, cm, tmpdir):
    tmp_ipynb = "notebook.ipynb"
    tmp_nbpy = "notebook.nb.py"

    cm.formats = "ipynb,.nb.py"
    cm.root_dir = str(tmpdir)

    # open ipynb, save nb.py, reopen
    nb = jupytext.read(ipynb_py_file)

    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_nbpy))
    nbpy = await ensure_async(cm.get(tmp_nbpy))
    compare_notebooks(nbpy["content"], nb)

    # open ipynb
    nbipynb = await ensure_async(cm.get(tmp_ipynb))
    compare_notebooks(nbipynb["content"], nb)

    # save ipynb
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # rename_file notebook.nb.py to new.nb.py
    await ensure_async(cm.rename_file(tmp_nbpy, "new.nb.py"))
    assert not os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert not os.path.isfile(str(tmpdir.join(tmp_nbpy)))

    assert os.path.isfile(str(tmpdir.join("new.ipynb")))
    assert os.path.isfile(str(tmpdir.join("new.nb.py")))

    # rename_file new.ipynb to notebook.ipynb
    await ensure_async(cm.rename_file("new.ipynb", tmp_ipynb))
    assert os.path.isfile(str(tmpdir.join(tmp_ipynb)))
    assert os.path.isfile(str(tmpdir.join(tmp_nbpy)))

    assert not os.path.isfile(str(tmpdir.join("new.ipynb")))
    assert not os.path.isfile(str(tmpdir.join("new.nb.py")))


async def test_load_save_rename_non_ascii_path(ipynb_py_file, cm, tmpdir):
    tmp_ipynb = "notebôk.ipynb"
    tmp_nbpy = "notebôk.nb.py"

    cm.formats = "ipynb,.nb.py"
    tmpdir = "" + str(tmpdir)
    cm.root_dir = tmpdir

    # open ipynb, save nb.py, reopen
    nb = jupytext.read(ipynb_py_file)

    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_nbpy))
    nbpy = await ensure_async(cm.get(tmp_nbpy))
    compare_notebooks(nbpy["content"], nb)

    # open ipynb
    nbipynb = await ensure_async(cm.get(tmp_ipynb))
    compare_notebooks(nbipynb["content"], nb)

    # save ipynb
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # rename_file notebôk.nb.py to nêw.nb.py
    await ensure_async(cm.rename_file(tmp_nbpy, "nêw.nb.py"))
    assert not os.path.isfile(os.path.join(tmpdir, tmp_ipynb))
    assert not os.path.isfile(os.path.join(tmpdir, tmp_nbpy))

    assert os.path.isfile(os.path.join(tmpdir, "nêw.ipynb"))
    assert os.path.isfile(os.path.join(tmpdir, "nêw.nb.py"))

    # rename_file nêw.ipynb to notebôk.ipynb
    await ensure_async(cm.rename_file("nêw.ipynb", tmp_ipynb))
    assert os.path.isfile(os.path.join(tmpdir, tmp_ipynb))
    assert os.path.isfile(os.path.join(tmpdir, tmp_nbpy))

    assert not os.path.isfile(os.path.join(tmpdir, "nêw.ipynb"))
    assert not os.path.isfile(os.path.join(tmpdir, "nêw.nb.py"))


async def test_outdated_text_notebook(python_notebook, cm, tmpdir):
    # 1. write py ipynb
    cm.formats = "py,ipynb"
    cm.outdated_text_notebook_margin = 0
    cm.root_dir = str(tmpdir)

    # open ipynb, save py, reopen
    nb = python_notebook
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.py"))
    model_py = await ensure_async(cm.get("notebook.py", load_alternative_format=False))
    model_ipynb = await ensure_async(
        cm.get("notebook.ipynb", load_alternative_format=False)
    )

    # 2. check that time of ipynb <= py
    assert model_ipynb["last_modified"] <= model_py["last_modified"]

    # 3. wait some time
    time.sleep(0.5)

    # 4. modify ipynb
    nb.cells.append(new_markdown_cell("New cell"))
    write(nb, str(tmpdir.join("notebook.ipynb")))

    # 5. test error
    with pytest.raises(HTTPError):
        await ensure_async(cm.get("notebook.py"))

    # 6. test OK with
    cm.outdated_text_notebook_margin = 1.0
    await ensure_async(cm.get("notebook.py"))

    # 7. test OK with
    cm.outdated_text_notebook_margin = float("inf")
    await ensure_async(cm.get("notebook.py"))


async def test_outdated_text_notebook_no_diff_ok(tmpdir, cm, python_notebook):
    # 1. write py ipynb
    cm.formats = "py,ipynb"
    cm.outdated_text_notebook_margin = 0
    cm.root_dir = str(tmpdir)

    # open ipynb, save py, reopen
    nb = python_notebook
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.py"))
    model_py = await ensure_async(cm.get("notebook.py", load_alternative_format=False))
    model_ipynb = await ensure_async(
        cm.get("notebook.ipynb", load_alternative_format=False)
    )

    # 2. check that time of ipynb <= py
    assert model_ipynb["last_modified"] <= model_py["last_modified"]

    # 3. wait some time
    time.sleep(0.5)

    # 4. touch ipynb
    with open(tmpdir / "notebook.ipynb", "a"):
        os.utime(tmpdir / "notebook.ipynb", None)

    # 5. No error since both files correspond to the same notebook #799
    await ensure_async(cm.get("notebook.py"))


async def test_outdated_text_notebook_diff_is_shown(tmpdir, cm, python_notebook):
    # 1. write py ipynb
    cm.formats = "py,ipynb"
    cm.outdated_text_notebook_margin = 0
    cm.root_dir = str(tmpdir)

    # open ipynb, save py, reopen
    nb = python_notebook
    nb.cells = [new_markdown_cell("Text version 1.0")]
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.py"))
    model_py = await ensure_async(cm.get("notebook.py", load_alternative_format=False))
    model_ipynb = await ensure_async(
        cm.get("notebook.ipynb", load_alternative_format=False)
    )

    # 2. check that time of ipynb <= py
    assert model_ipynb["last_modified"] <= model_py["last_modified"]

    # 3. wait some time
    time.sleep(0.5)

    # 4. modify ipynb
    nb.cells = [new_markdown_cell("Text version 2.0")]
    jupytext.write(nb, str(tmpdir / "notebook.ipynb"))

    # 5. The diff is shown in the error
    with pytest.raises(HTTPError) as excinfo:
        await ensure_async(cm.get("notebook.py"))

    diff = excinfo.value.log_message

    diff = diff[diff.find("Differences") : diff.rfind("Please")]

    compare(
        # In the reference below, lines with a single space
        # have been stripped by the pre-commit hook
        diff.replace("\n \n", "\n\n"),
        """Differences (jupytext --diff notebook.py notebook.ipynb) are:
--- notebook.py
+++ notebook.ipynb
@@ -13,5 +13,5 @@
 # ---

 # %%%% [markdown]
-# Text version 1.0
+# Text version 2.0

""",
    )


async def test_reload_notebook_after_jupytext_cli(python_notebook, cm, tmpdir):
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_nbpy = str(tmpdir.join("notebook.py"))

    cm.outdated_text_notebook_margin = 0
    cm.root_dir = str(tmpdir)

    # write the paired notebook
    nb = python_notebook
    nb.metadata.setdefault("jupytext", {})["formats"] = "py,ipynb"
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.py"))

    assert os.path.isfile(tmp_ipynb)
    assert os.path.isfile(tmp_nbpy)

    # run jupytext CLI
    jupytext_cli([tmp_nbpy, "--to", "ipynb", "--update"])

    # test reload
    nb1 = (await ensure_async(cm.get("notebook.py")))["content"]
    nb2 = (await ensure_async(cm.get("notebook.ipynb")))["content"]

    compare_notebooks(nb, nb1)
    compare_notebooks(nb, nb2)


async def test_load_save_percent_format(percent_file, cm, tmpdir):
    tmp_py = "notebook.py"
    with open(percent_file) as stream:
        text_py = stream.read()
    with open(str(tmpdir.join(tmp_py)), "w") as stream:
        stream.write(text_py)

    cm.root_dir = str(tmpdir)

    # open python, save
    nb = (await ensure_async(cm.get(tmp_py)))["content"]
    del nb.metadata["jupytext"]["notebook_metadata_filter"]
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_py))

    # compare the new file with original one
    with open(str(tmpdir.join(tmp_py))) as stream:
        text_py2 = stream.read()

    # do we find 'percent' in the header?
    header = text_py2[: -len(text_py)]
    assert any(["percent" in line for line in header.splitlines()])

    # Remove the YAML header
    text_py2 = text_py2[-len(text_py) :]

    compare(text_py2, text_py)


async def test_save_to_percent_format(ipynb_julia_file, cm, tmpdir):
    tmp_ipynb = "notebook.ipynb"
    tmp_jl = "notebook.jl"

    cm.root_dir = str(tmpdir)
    cm.preferred_jupytext_formats_save = "jl:percent"

    nb = jupytext.read(ipynb_julia_file)
    nb["metadata"]["jupytext"] = {"formats": "ipynb,jl"}

    # save to ipynb and jl
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # read jl file
    with open(str(tmpdir.join(tmp_jl))) as stream:
        text_jl = stream.read()

    # Parse the YAML header
    metadata, _, _, _ = header_to_metadata_and_cell(text_jl.splitlines(), "#", "")
    assert metadata["jupytext"]["formats"] == "ipynb,jl:percent"


async def test_save_using_preferred_and_default_format_170(ipynb_py_file, cm, tmpdir):
    nb = read(ipynb_py_file)

    # Way 0: preferred_jupytext_formats_save, no prefix + formats
    tmp_py = str(tmpdir.join("python/notebook.py"))

    cm.root_dir = str(tmpdir)
    cm.preferred_jupytext_formats_save = "py:percent"
    cm.formats = "ipynb,python//py"

    # save to ipynb and py
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.ipynb"))

    # read py file
    nb_py = read(tmp_py)
    assert nb_py.metadata["jupytext"]["text_representation"]["format_name"] == "percent"

    # Way 1: preferred_jupytext_formats_save + formats
    tmp_py = str(tmpdir.join("python/notebook.py"))

    cm.root_dir = str(tmpdir)
    cm.preferred_jupytext_formats_save = "python//py:percent"
    cm.formats = "ipynb,python//py"

    # save to ipynb and py
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.ipynb"))

    # read py file
    nb_py = read(tmp_py)
    assert nb_py.metadata["jupytext"]["text_representation"]["format_name"] == "percent"

    # Way 2: formats
    tmp_py = str(tmpdir.join("python/notebook.py"))

    cm.root_dir = str(tmpdir)
    cm.formats = "ipynb,python//py:percent"

    # save to ipynb and py
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.ipynb"))

    # read py file
    nb_py = read(tmp_py)
    assert nb_py.metadata["jupytext"]["text_representation"]["format_name"] == "percent"


async def test_open_using_preferred_and_default_format_174(ipynb_py_file, cm, tmpdir):
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("python/notebook.py"))
    tmp_py2 = str(tmpdir.join("other/notebook.py"))
    os.makedirs(str(tmpdir.join("other")))
    shutil.copyfile(ipynb_py_file, tmp_ipynb)

    cm.root_dir = str(tmpdir)
    cm.formats = "ipynb,python//py:percent"
    cm.notebook_metadata_filter = "all"
    cm.cell_metadata_filter = "all"

    # load notebook
    model = await ensure_async(cm.get("notebook.ipynb"))

    # save to ipynb and py
    await ensure_async(cm.save(model=model, path="notebook.ipynb"))

    assert os.path.isfile(tmp_py)
    os.remove(tmp_ipynb)

    # read py file
    model2 = await ensure_async(cm.get("python/notebook.py"))
    compare_notebooks(model2["content"], model["content"])

    # move py file to the another folder
    shutil.move(tmp_py, tmp_py2)
    model2 = await ensure_async(cm.get("other/notebook.py"))
    compare_notebooks(model2["content"], model["content"])
    await ensure_async(cm.save(model=model, path="other/notebook.py"))
    assert not os.path.isfile(tmp_ipynb)
    assert not os.path.isfile(str(tmpdir.join("other/notebook.ipynb")))


async def test_kernelspec_are_preserved(ipynb_py_file, cm, tmpdir):
    if "many hash" in ipynb_py_file:
        pytest.skip()
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("notebook.py"))
    shutil.copyfile(ipynb_py_file, tmp_ipynb)

    cm.root_dir = str(tmpdir)
    cm.formats = "ipynb,py"
    cm.notebook_metadata_filter = "-all"

    # load notebook
    model = await ensure_async(cm.get("notebook.ipynb"))
    model["content"].metadata["kernelspec"] = {
        "display_name": "Kernel name",
        "language": "python",
        "name": "custom",
    }

    # save to ipynb and py
    await ensure_async(cm.save(model=model, path="notebook.ipynb"))
    assert os.path.isfile(tmp_py)

    # read ipynb
    model2 = await ensure_async(cm.get("notebook.ipynb"))
    compare_notebooks(model2["content"], model["content"])


async def test_save_to_light_percent_sphinx_format(ipynb_py_file, cm, tmpdir):
    tmp_ipynb = "notebook.ipynb"
    tmp_lgt_py = "notebook.lgt.py"
    tmp_pct_py = "notebook.pct.py"
    tmp_spx_py = "notebook.spx.py"

    cm.root_dir = str(tmpdir)

    nb = jupytext.read(ipynb_py_file)
    nb["metadata"]["jupytext"] = {
        "formats": "ipynb,.pct.py:percent,.lgt.py:light,.spx.py:sphinx"
    }

    # save to ipynb and three python flavors
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # read files
    with open(str(tmpdir.join(tmp_pct_py))) as stream:
        assert read_format_from_metadata(stream.read(), ".py") == "percent"

    with open(str(tmpdir.join(tmp_lgt_py))) as stream:
        assert read_format_from_metadata(stream.read(), ".py") == "light"

    with open(str(tmpdir.join(tmp_spx_py))) as stream:
        assert read_format_from_metadata(stream.read(), ".py") == "sphinx"

    model = await ensure_async(cm.get(path=tmp_pct_py))
    compare_notebooks(model["content"], nb)

    model = await ensure_async(cm.get(path=tmp_lgt_py))
    compare_notebooks(model["content"], nb)

    model = await ensure_async(cm.get(path=tmp_spx_py))
    # (notebooks not equal as we insert %matplotlib inline in sphinx)

    model = await ensure_async(cm.get(path=tmp_ipynb))
    compare_notebooks(model["content"], nb)


async def test_pair_notebook_with_dot(ipynb_py_file, cm, tmpdir):
    # Reproduce issue #138
    tmp_py = "file.5.1.py"
    tmp_ipynb = "file.5.1.ipynb"

    cm.root_dir = str(tmpdir)

    nb = jupytext.read(ipynb_py_file)
    nb["metadata"]["jupytext"] = {"formats": "ipynb,py:percent"}

    # save to ipynb and three python flavors
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    assert os.path.isfile(str(tmpdir.join(tmp_ipynb)))

    # read files
    with open(str(tmpdir.join(tmp_py))) as stream:
        assert read_format_from_metadata(stream.read(), ".py") == "percent"

    model = await ensure_async(cm.get(path=tmp_py))
    assert model["name"] == "file.5.1.py"
    compare_notebooks(model["content"], nb)

    model = await ensure_async(cm.get(path=tmp_ipynb))
    assert model["name"] == "file.5.1.ipynb"
    compare_notebooks(model["content"], nb)


async def test_preferred_format_allows_to_read_others_format(
    python_notebook, cm, tmpdir
):
    # 1. write py ipynb
    tmp_ipynb = "notebook.ipynb"
    tmp_nbpy = "notebook.py"

    cm.preferred_jupytext_formats_save = "py:light"
    cm.root_dir = str(tmpdir)

    # load notebook and save it using the cm
    nb = python_notebook
    nb["metadata"]["jupytext"] = {"formats": "ipynb,py"}
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # Saving does not update the metadata, as 'save' makes a copy of the notebook
    # assert nb['metadata']['jupytext']['formats'] == 'ipynb,py:light'

    # Set preferred format for reading
    cm.preferred_jupytext_formats_read = "py:percent"

    # Read notebook
    model = await ensure_async(cm.get(tmp_nbpy))

    # Check that format is explicit
    assert model["content"]["metadata"]["jupytext"]["formats"] == "ipynb,py:light"

    # Check contents
    compare_notebooks(model["content"], nb)

    # Change save format and save
    model["content"]["metadata"]["jupytext"]["formats"] == "ipynb,py"
    cm.preferred_jupytext_formats_save = "py:percent"
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # Read notebook
    model = await ensure_async(cm.get(tmp_nbpy))
    compare_notebooks(model["content"], nb)

    # Check that format is explicit
    assert model["content"]["metadata"]["jupytext"]["formats"] == "ipynb,py:percent"


async def test_preferred_formats_read_auto(tmpdir, cm):
    tmp_py = "notebook.py"
    with open(str(tmpdir.join(tmp_py)), "w") as script:
        script.write(
            """# cell one
1 + 1
"""
        )

    # create contents manager with default load format as percent
    cm.preferred_jupytext_formats_read = "auto:percent"
    cm.root_dir = str(tmpdir)

    # load notebook
    model = await ensure_async(cm.get(tmp_py))

    # check that script is opened as percent
    assert (
        "percent"
        == model["content"]["metadata"]["jupytext"]["text_representation"][
            "format_name"
        ]
    )


async def test_save_in_auto_extension_global(ipynb_py_R_jl_file, cm, tmpdir):
    # load notebook
    nb = jupytext.read(ipynb_py_R_jl_file)

    auto_ext = auto_ext_from_metadata(nb.metadata)
    tmp_ipynb = "notebook.ipynb"
    tmp_script = "notebook" + auto_ext

    # create contents manager with default load format as percent
    cm.formats = "ipynb,auto"
    cm.preferred_jupytext_formats_save = "auto:percent"
    cm.root_dir = str(tmpdir)

    # save notebook
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # check that text representation exists, and is in percent format
    with open(str(tmpdir.join(tmp_script))) as stream:
        assert read_format_from_metadata(stream.read(), auto_ext) == "percent"

    # reload and compare with original notebook
    model = await ensure_async(cm.get(path=tmp_script))

    # saving should not create a format entry #95
    assert "formats" not in model["content"].metadata.get("jupytext", {})

    compare_notebooks(model["content"], nb)


async def test_global_auto_pairing_works_with_empty_notebook(tmpdir, cm):
    nb = new_notebook()
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("notebook.py"))
    tmp_auto = str(tmpdir.join("notebook.auto"))

    # create contents manager with default load format as percent
    cm.formats = "ipynb,auto"
    cm.preferred_jupytext_formats_save = "auto:percent"
    cm.root_dir = str(tmpdir)

    # save notebook
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.ipynb"))

    # check that only the ipynb representation exists
    assert os.path.isfile(tmp_ipynb)
    assert not os.path.isfile(tmp_py)
    assert not os.path.isfile(tmp_auto)
    assert "notebook.ipynb" not in cm.paired_notebooks

    model = await ensure_async(cm.get(path="notebook.ipynb"))
    compare_notebooks(model["content"], nb)

    # add language information to the notebook
    nb.metadata["language_info"] = {
        "codemirror_mode": {"name": "ipython", "version": 3},
        "file_extension": ".py",
        "mimetype": "text/x-python",
        "name": "python",
        "nbconvert_exporter": "python",
        "pygments_lexer": "ipython3",
        "version": "3.7.3",
    }

    # save again
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.ipynb"))

    # check that ipynb + py representations exists
    assert os.path.isfile(tmp_ipynb)
    assert os.path.isfile(tmp_py)
    assert not os.path.isfile(tmp_auto)
    assert len(cm.paired_notebooks["notebook.ipynb"]) == 2

    # add a cell in the py file
    with open(tmp_py, "a") as fp:
        fp.write("# %%\n2+2\n")

    nb2 = (await ensure_async(cm.get(path="notebook.ipynb")))["content"]
    assert len(nb2.cells) == 1
    assert nb2.cells[0].source == "2+2"


async def test_save_in_auto_extension_global_with_format(
    ipynb_py_R_jl_file, cm, tmpdir
):
    # load notebook
    nb = jupytext.read(ipynb_py_R_jl_file)

    auto_ext = auto_ext_from_metadata(nb.metadata)
    tmp_ipynb = "notebook.ipynb"
    tmp_script = "notebook" + auto_ext

    # create contents manager with default load format as percent
    cm.formats = "ipynb,auto:percent"
    cm.root_dir = str(tmpdir)

    # save notebook
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # check that text representation exists, and is in percent format
    with open(str(tmpdir.join(tmp_script))) as stream:
        assert read_format_from_metadata(stream.read(), auto_ext) == "percent"

    # reload and compare with original notebook
    model = await ensure_async(cm.get(path=tmp_script))

    # saving should not create a format entry #95
    assert "formats" not in model["content"].metadata.get("jupytext", {})

    compare_notebooks(model["content"], nb)


async def test_save_in_auto_extension_local(ipynb_py_R_jl_file, cm, tmpdir):
    # load notebook
    nb = jupytext.read(ipynb_py_R_jl_file)
    nb.metadata.setdefault("jupytext", {})["formats"] = "ipynb,auto:percent"

    auto_ext = auto_ext_from_metadata(nb.metadata)
    tmp_ipynb = "notebook.ipynb"
    tmp_script = "notebook" + auto_ext

    # create contents manager with default load format as percent
    cm.root_dir = str(tmpdir)

    # save notebook
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # check that text representation exists, and is in percent format
    with open(str(tmpdir.join(tmp_script))) as stream:
        assert read_format_from_metadata(stream.read(), auto_ext) == "percent"

    # reload and compare with original notebook
    model = await ensure_async(cm.get(path=tmp_script))

    compare_notebooks(model["content"], nb)


async def test_save_in_pct_and_lgt_auto_extensions(ipynb_py_R_jl_file, cm, tmpdir):
    # load notebook
    nb = jupytext.read(ipynb_py_R_jl_file)

    auto_ext = auto_ext_from_metadata(nb.metadata)
    tmp_ipynb = "notebook.ipynb"
    tmp_pct_script = "notebook.pct" + auto_ext
    tmp_lgt_script = "notebook.lgt" + auto_ext

    # create contents manager with default load format as percent
    cm.formats = "ipynb,.pct.auto,.lgt.auto"
    cm.preferred_jupytext_formats_save = ".pct.auto:percent,.lgt.auto:light"
    cm.root_dir = str(tmpdir)

    # save notebook
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # check that text representation exists in percent format
    with open(str(tmpdir.join(tmp_pct_script))) as stream:
        assert read_format_from_metadata(stream.read(), auto_ext) == "percent"

    # check that text representation exists in light format
    with open(str(tmpdir.join(tmp_lgt_script))) as stream:
        assert read_format_from_metadata(stream.read(), auto_ext) == "light"


async def test_metadata_filter_is_effective(ipynb_py_R_jl_file, cm, tmpdir):
    if re.match(r".*(magic|305).*", ipynb_py_R_jl_file):
        pytest.skip()
    nb = jupytext.read(ipynb_py_R_jl_file)
    tmp_ipynb = "notebook.ipynb"
    tmp_script = "notebook.py"

    # create contents manager
    cm.root_dir = str(tmpdir)

    # save notebook to tmpdir
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # set config
    cm.formats = "ipynb,py"
    cm.notebook_metadata_filter = "jupytext,-all"
    cm.cell_metadata_filter = "-all"

    # load notebook
    nb = (await ensure_async(cm.get(tmp_ipynb)))["content"]

    assert nb.metadata["jupytext"]["cell_metadata_filter"] == "-all"
    assert nb.metadata["jupytext"]["notebook_metadata_filter"] == "jupytext,-all"

    # save notebook again
    await ensure_async(cm.save(model=notebook_model(nb), path=tmp_ipynb))

    # read text version
    nb2 = jupytext.read(str(tmpdir.join(tmp_script)))

    # test no metadata
    assert set(nb2.metadata.keys()) <= {"jupytext", "kernelspec"}
    for cell in nb2.cells:
        assert not cell.metadata

    # read paired notebook
    nb3 = (await ensure_async(cm.get(tmp_script)))["content"]

    compare_notebooks(nb3, nb)


async def test_no_metadata_added_to_scripts_139(tmpdir, cm):
    tmp_script = str(tmpdir.join("script.py"))
    text = """import os


print('hello1')



print('hello2')
"""

    with open(tmp_script, "w") as fp:
        fp.write(text)

    # create contents manager
    cm.root_dir = str(tmpdir)

    # Andre's config #139
    cm.freeze_metadata = True
    cm.notebook_metadata_filter = "-all"
    cm.cell_metadata_filter = "-lines_to_next_cell"

    # load notebook
    model = await ensure_async(cm.get("script.py"))

    # add cell metadata
    for cell in model["content"].cells:
        cell.metadata.update(
            {
                "ExecuteTime": {
                    "start_time": "2019-02-06T11:53:21.208644Z",
                    "end_time": "2019-02-06T11:53:21.213071Z",
                }
            }
        )

    # save notebook
    await ensure_async(cm.save(model=model, path="script.py"))

    with open(tmp_script) as fp:
        compare(fp.read(), text)


@pytest.mark.parametrize("ext", [".py", ".ipynb"])
async def test_local_format_can_deactivate_pairing(ipynb_py_file, cm, ext, tmpdir):
    """This is a test for #157: local format can be used to deactivate the global pairing"""
    nb = jupytext.read(ipynb_py_file)
    nb.metadata["jupytext_formats"] = ext[1:]  # py or ipynb

    # create contents manager with default pairing
    cm.formats = "ipynb,py"
    cm.root_dir = str(tmpdir)

    # save notebook
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook" + ext))

    # check that only the text representation exists
    assert os.path.isfile(str(tmpdir.join("notebook.py"))) == (ext == ".py")
    assert os.path.isfile(str(tmpdir.join("notebook.ipynb"))) == (ext == ".ipynb")
    nb2 = (await ensure_async(cm.get("notebook" + ext)))["content"]
    compare_notebooks(nb2, nb)

    # resave, check again
    await ensure_async(cm.save(model=notebook_model(nb2), path="notebook" + ext))

    assert os.path.isfile(str(tmpdir.join("notebook.py"))) == (ext == ".py")
    assert os.path.isfile(str(tmpdir.join("notebook.ipynb"))) == (ext == ".ipynb")
    nb3 = (await ensure_async(cm.get("notebook" + ext)))["content"]
    compare_notebooks(nb3, nb)


async def test_global_pairing_allows_to_save_other_file_types(rmd_file, cm, tmpdir):
    """This is a another test for #157: local format can be used to deactivate the global pairing"""
    nb = jupytext.read(rmd_file)

    # create contents manager with default pairing
    cm.formats = "ipynb,py"
    cm.root_dir = str(tmpdir)

    # save notebook
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.Rmd"))

    # check that only the original file is saved
    assert os.path.isfile(str(tmpdir.join("notebook.Rmd")))
    assert not os.path.isfile(str(tmpdir.join("notebook.py")))
    assert not os.path.isfile(str(tmpdir.join("notebook.ipynb")))

    nb2 = (await ensure_async(cm.get("notebook.Rmd")))["content"]
    compare_notebooks(nb2, nb)


@pytest.mark.requires_user_kernel_python3
async def test_python_kernel_preserves_R_files(r_file, cm, tmpdir):
    """Opening a R file with a Jupyter server that has no R kernel should not modify the file"""
    tmp_r_file = str(tmpdir.join("script.R"))
    with open(r_file) as fp:
        script = fp.read()
    with open(tmp_r_file, "w") as fp:
        fp.write(script)

    # create contents manager
    cm.root_dir = str(tmpdir)

    # open notebook, set Python kernel and save
    model = await ensure_async(cm.get("script.R"))
    model["content"].metadata["kernelspec"] = kernelspec_from_language("python")
    await ensure_async(cm.save(model=model, path="script.R"))

    with open(tmp_r_file) as fp:
        script2 = fp.read()

    compare(script2, script)


async def test_pair_notebook_in_another_folder(tmpdir, cm):
    cm.root_dir = str(tmpdir)

    os.makedirs(str(tmpdir.join("notebooks")))
    tmp_ipynb = str(tmpdir.join("notebooks/notebook_name.ipynb"))
    tmp_py = str(tmpdir.join("scripts/notebook_name.py"))

    await ensure_async(
        cm.save(
            model=notebook_model(
                new_notebook(
                    metadata={"jupytext": {"formats": "notebooks//ipynb,scripts//py"}}
                ),
            ),
            path="notebooks/notebook_name.ipynb",
        )
    )

    assert os.path.isfile(tmp_ipynb)
    assert os.path.isfile(tmp_py)

    await ensure_async(cm.get("notebooks/notebook_name.ipynb"))
    await ensure_async(cm.get("scripts/notebook_name.py"))


async def test_pair_notebook_in_dotdot_folder(tmpdir, cm):
    cm.root_dir = str(tmpdir)

    os.makedirs(str(tmpdir.join("notebooks")))
    tmp_ipynb = str(tmpdir.join("notebooks/notebook_name.ipynb"))
    tmp_py = str(tmpdir.join("scripts/notebook_name.py"))

    await ensure_async(
        cm.save(
            model=notebook_model(
                new_notebook(metadata={"jupytext": {"formats": "ipynb,../scripts//py"}})
            ),
            path="notebooks/notebook_name.ipynb",
        )
    )

    assert os.path.isfile(tmp_ipynb)
    assert os.path.isfile(tmp_py)

    await ensure_async(cm.get("notebooks/notebook_name.ipynb"))
    await ensure_async(cm.get("scripts/notebook_name.py"))


async def test_split_at_heading_option(tmpdir, cm):
    text = """Markdown text

# Header one

## Header two
"""
    tmp_md = str(tmpdir.join("notebook.md"))
    with open(tmp_md, "w") as fp:
        fp.write(text)

    cm.root_dir = str(tmpdir)
    cm.split_at_heading = True

    nb = (await ensure_async(cm.get("notebook.md")))["content"]

    # Was rst to md conversion effective?
    assert nb.cells[0].source == "Markdown text"
    assert nb.cells[1].source == "# Header one"
    assert nb.cells[2].source == "## Header two"

    nb.metadata["jupytext"]["notebook_metadata_filter"] = "-all"
    text2 = writes(nb, "md")
    compare(text2, text)


async def test_load_then_change_formats(tmpdir, cm):
    tmp_ipynb = str(tmpdir.join("nb.ipynb"))
    tmp_py = str(tmpdir.join("nb.py"))
    nb = new_notebook(metadata={"jupytext": {"formats": "ipynb,py:light"}})
    write(nb, tmp_ipynb)

    cm.root_dir = str(tmpdir)

    model = await ensure_async(cm.get("nb.ipynb"))
    assert model["content"].metadata["jupytext"]["formats"] == "ipynb,py:light"

    await ensure_async(cm.save(model, path="nb.ipynb"))
    assert os.path.isfile(tmp_py)
    assert read(tmp_py).metadata["jupytext"]["formats"] == "ipynb,py:light"

    time.sleep(0.5)
    del model["content"].metadata["jupytext"]["formats"]
    await ensure_async(cm.save(model, path="nb.ipynb"))
    # test that we have not kept the 'ipynb/py' pairing info, and that we can read the ipynb
    await ensure_async(cm.get("nb.ipynb"))
    os.remove(tmp_py)

    model["content"].metadata.setdefault("jupytext", {})["formats"] = "ipynb,py:percent"
    await ensure_async(cm.save(model, path="nb.ipynb"))
    assert os.path.isfile(tmp_py)
    assert read(tmp_py).metadata["jupytext"]["formats"] == "ipynb,py:percent"
    os.remove(tmp_py)

    del model["content"].metadata["jupytext"]["formats"]
    await ensure_async(cm.save(model, path="nb.ipynb"))
    assert not os.path.isfile(tmp_py)


async def test_set_then_change_formats(tmpdir, cm):
    tmp_py = str(tmpdir.join("nb.py"))
    nb = new_notebook(metadata={"jupytext": {"formats": "ipynb,py:light"}})

    cm.root_dir = str(tmpdir)

    await ensure_async(cm.save(model=notebook_model(nb), path="nb.ipynb"))
    assert os.path.isfile(tmp_py)
    assert read(tmp_py).metadata["jupytext"]["formats"] == "ipynb,py:light"
    os.remove(tmp_py)

    nb.metadata["jupytext"]["formats"] = "ipynb,py:percent"
    await ensure_async(cm.save(model=notebook_model(nb), path="nb.ipynb"))
    assert os.path.isfile(tmp_py)
    assert read(tmp_py).metadata["jupytext"]["formats"] == "ipynb,py:percent"
    os.remove(tmp_py)

    del nb.metadata["jupytext"]["formats"]
    await ensure_async(cm.save(model=notebook_model(nb), path="nb.ipynb"))
    assert not os.path.isfile(tmp_py)


async def test_set_then_change_auto_formats(tmpdir, cm, python_notebook):
    tmp_ipynb = str(tmpdir.join("nb.ipynb"))
    tmp_py = str(tmpdir.join("nb.py"))
    tmp_rmd = str(tmpdir.join("nb.Rmd"))
    nb = new_notebook(metadata=python_notebook.metadata)

    cm.root_dir = str(tmpdir)

    # Pair ipynb/py and save
    nb.metadata["jupytext"] = {"formats": "ipynb,auto:light"}
    await ensure_async(cm.save(model=notebook_model(nb), path="nb.ipynb"))
    assert "nb.py" in cm.paired_notebooks
    assert "nb.auto" not in cm.paired_notebooks
    assert os.path.isfile(tmp_py)
    assert read(tmp_ipynb).metadata["jupytext"]["formats"] == "ipynb,py:light"

    # Pair ipynb/Rmd and save
    time.sleep(0.5)
    nb.metadata["jupytext"] = {"formats": "ipynb,Rmd"}
    await ensure_async(cm.save(model=notebook_model(nb), path="nb.ipynb"))
    assert "nb.Rmd" in cm.paired_notebooks
    assert "nb.py" not in cm.paired_notebooks
    assert "nb.auto" not in cm.paired_notebooks
    assert os.path.isfile(tmp_rmd)
    assert read(tmp_ipynb).metadata["jupytext"]["formats"] == "ipynb,Rmd"
    await ensure_async(cm.get("nb.ipynb"))

    # Unpair and save
    time.sleep(0.5)
    del nb.metadata["jupytext"]
    await ensure_async(cm.save(model=notebook_model(nb), path="nb.ipynb"))
    assert "nb.Rmd" not in cm.paired_notebooks
    assert "nb.py" not in cm.paired_notebooks
    assert "nb.auto" not in cm.paired_notebooks
    await ensure_async(cm.get("nb.ipynb"))


async def test_share_py_recreate_ipynb(tmpdir, cm, ipynb_py_R_jl_file):
    tmp_ipynb = str(tmpdir.join("nb.ipynb"))
    tmp_py = str(tmpdir.join("nb.py"))

    cm.root_dir = str(tmpdir)

    # set default py format
    cm.preferred_jupytext_formats_save = "py:percent"

    # every new file is paired
    cm.formats = "ipynb,py"

    # the text files don't need a YAML header
    cm.notebook_metadata_filter = "-all"
    cm.cell_metadata_filter = "-all"

    nb = read(ipynb_py_R_jl_file)
    model_ipynb = await ensure_async(cm.save(model=notebook_model(nb), path="nb.ipynb"))

    assert os.path.isfile(tmp_ipynb)
    assert os.path.isfile(tmp_py)

    os.remove(tmp_ipynb)

    # reopen and save nb.py
    model = await ensure_async(cm.get("nb.py"))
    await ensure_async(cm.save(model=model, path="nb.py"))

    # ipynb is re-created
    assert os.path.isfile(tmp_ipynb)

    # save time of ipynb is that of py file
    assert model_ipynb["last_modified"] == model["last_modified"]


async def test_vim_folding_markers(tmpdir, cm):
    tmp_ipynb = str(tmpdir.join("nb.ipynb"))
    tmp_py = str(tmpdir.join("nb.py"))

    cm.root_dir = str(tmpdir)

    # Default Vim folding markers
    cm.cell_markers = "{{{,}}}"
    cm.formats = "ipynb,py:light"

    nb = new_notebook(
        cells=[
            new_code_cell(
                """# region
'''Sample cell with region markers'''
'''End of the cell'''
# end region"""
            ),
            new_code_cell("a = 1\n\n\nb = 1"),
        ]
    )
    await ensure_async(cm.save(model=notebook_model(nb), path="nb.ipynb"))

    assert os.path.isfile(tmp_ipynb)
    assert os.path.isfile(tmp_py)

    nb2 = (await ensure_async(cm.get("nb.ipynb")))["content"]
    compare_notebooks(nb2, nb)

    nb3 = read(tmp_py)
    assert nb3.metadata["jupytext"]["cell_markers"] == "{{{,}}}"

    with open(tmp_py) as fp:
        text = fp.read()

    # Remove YAML header
    text = re.sub(re.compile(r"# ---.*# ---\n\n", re.DOTALL), "", text)

    compare(
        text,
        """# region
'''Sample cell with region markers'''
'''End of the cell'''
# end region

# {{{
a = 1


b = 1
# }}}
""",
    )


async def test_vscode_pycharm_folding_markers(tmpdir, cm):
    tmp_ipynb = str(tmpdir.join("nb.ipynb"))
    tmp_py = str(tmpdir.join("nb.py"))

    cm.root_dir = str(tmpdir)

    # Default VScode/PyCharm folding markers
    cm.cell_markers = "region,endregion"
    cm.formats = "ipynb,py:light"

    nb = new_notebook(
        cells=[
            new_code_cell(
                """# {{{
'''Sample cell with region markers'''
'''End of the cell'''
# }}}"""
            ),
            new_code_cell("a = 1\n\n\nb = 1"),
        ]
    )
    await ensure_async(cm.save(model=notebook_model(nb), path="nb.ipynb"))

    assert os.path.isfile(tmp_ipynb)
    assert os.path.isfile(tmp_py)

    nb2 = (await ensure_async(cm.get("nb.ipynb")))["content"]
    compare_notebooks(nb2, nb)

    nb3 = read(tmp_py)
    assert nb3.metadata["jupytext"]["cell_markers"] == "region,endregion"

    with open(tmp_py) as fp:
        text = fp.read()

    # Remove YAML header
    text = re.sub(re.compile(r"# ---.*# ---\n\n", re.DOTALL), "", text)

    compare(
        text,
        """# {{{
'''Sample cell with region markers'''
'''End of the cell'''
# }}}

# region
a = 1


b = 1
# endregion
""",
    )


async def test_open_file_with_cell_markers(tmpdir, cm):
    tmp_py = str(tmpdir.join("nb.py"))

    cm.root_dir = str(tmpdir)

    # Default VScode/PyCharm folding markers
    cm.cell_markers = "region,endregion"

    text = """# +
# this is a unique code cell
1 + 1

2 + 2
"""

    with open(tmp_py, "w") as fp:
        fp.write(text)

    nb = (await ensure_async(cm.get("nb.py")))["content"]
    assert len(nb.cells) == 1

    await ensure_async(cm.save(model=notebook_model(nb), path="nb.py"))

    with open(tmp_py) as fp:
        text2 = fp.read()

    expected = """# region
# this is a unique code cell
1 + 1

2 + 2
# endregion
"""

    compare(text2, expected)


async def test_save_file_with_cell_markers(tmpdir, cm):
    tmp_py = str(tmpdir.join("nb.py"))

    cm.root_dir = str(tmpdir)

    # Default VScode/PyCharm folding markers
    cm.cell_markers = "region,endregion"

    text = """# +
# this is a unique code cell
1 + 1

2 + 2
"""

    with open(tmp_py, "w") as fp:
        fp.write(text)

    nb = (await ensure_async(cm.get("nb.py")))["content"]
    assert len(nb.cells) == 1

    await ensure_async(cm.save(model=notebook_model(nb), path="nb.py"))

    with open(tmp_py) as fp:
        text2 = fp.read()

    compare(
        text2,
        """# region
# this is a unique code cell
1 + 1

2 + 2
# endregion
""",
    )

    nb2 = (await ensure_async(cm.get("nb.py")))["content"]
    compare_notebooks(nb2, nb)
    assert nb2.metadata["jupytext"]["cell_markers"] == "region,endregion"


async def test_notebook_extensions(tmpdir, cm, cwd_tmpdir):
    nb = new_notebook()
    write(nb, "script.py")
    write(nb, "notebook.Rmd")
    write(nb, "notebook.ipynb")

    cm.root_dir = str(tmpdir)
    cm.notebook_extensions = "ipynb,Rmd"

    model = await ensure_async(cm.get("notebook.ipynb"))
    assert model["type"] == "notebook"

    model = await ensure_async(cm.get("notebook.Rmd"))
    assert model["type"] == "notebook"

    model = await ensure_async(cm.get("script.py"))
    assert model["type"] == "file"


async def test_notebook_extensions_in_config(tmpdir, cm, cwd_tmpdir):
    nb = new_notebook()
    write(nb, "script.py")
    write(nb, "notebook.Rmd")
    write(nb, "notebook.ipynb")
    tmpdir.join("jupytext.toml").write("""notebook_extensions = ["ipynb", "Rmd"]""")

    cm.root_dir = str(tmpdir)

    model = await ensure_async(cm.get("notebook.ipynb"))
    assert model["type"] == "notebook"

    model = await ensure_async(cm.get("notebook.Rmd"))
    assert model["type"] == "notebook"

    model = await ensure_async(cm.get("script.py"))
    assert model["type"] == "file"


async def test_invalid_config_in_cm(tmpdir, cm, cwd_tmpdir):
    nb = new_notebook()
    write(nb, "notebook.ipynb")
    tmpdir.join("pyproject.toml").write(
        """[tool.jupysql.SqlMagic]
autopandas = False
displaylimit = 1"""
    )

    cm.root_dir = str(tmpdir)

    # list directory
    await ensure_async(cm.get(""))

    model = await ensure_async(cm.get("notebook.ipynb"))
    assert model["type"] == "notebook"


async def test_download_file_318(tmpdir, cm):
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("notebook.py"))

    nb = new_notebook()
    nb.metadata["jupytext"] = {"formats": "ipynb,py"}
    write(nb, tmp_ipynb)
    write(nb, tmp_py)

    cm.root_dir = str(tmpdir)
    cm.notebook_extensions = "ipynb"

    model = await ensure_async(
        cm.get("notebook.ipynb", content=True, type=None, format=None)
    )
    assert model["type"] == "notebook"


async def test_markdown_and_r_extensions(tmpdir, cm):
    tmp_r = str(tmpdir.join("script.r"))
    tmp_markdown = str(tmpdir.join("notebook.markdown"))

    nb = new_notebook()
    write(nb, tmp_r)
    write(nb, tmp_markdown)

    cm.root_dir = str(tmpdir)

    model = await ensure_async(cm.get("script.r"))
    assert model["type"] == "notebook"

    model = await ensure_async(cm.get("notebook.markdown"))
    assert model["type"] == "notebook"


async def test_server_extension_issubclass(cm):
    class SubClassTextFileContentsManager(jupytext.TextFileContentsManager):
        pass

    assert not isinstance(
        SubClassTextFileContentsManager, jupytext.TextFileContentsManager
    )
    assert issubclass(SubClassTextFileContentsManager, jupytext.TextFileContentsManager)


async def test_multiple_pairing(tmpdir, cm):
    """Test that multiple pairing works. Input cells are loaded from the most recent text representation among
    the paired ones"""
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_md = str(tmpdir.join("notebook.md"))
    tmp_py = str(tmpdir.join("notebook.py"))

    def nb(text):
        return new_notebook(
            cells=[new_markdown_cell(text)],
            metadata={"jupytext": {"formats": "ipynb,md,py"}},
        )

    cm.root_dir = str(tmpdir)

    await ensure_async(
        cm.save(model=notebook_model(nb("saved from cm")), path="notebook.ipynb")
    )
    compare_notebooks(jupytext.read(tmp_ipynb), nb("saved from cm"))
    compare_notebooks(jupytext.read(tmp_md), nb("saved from cm"))
    compare_notebooks(jupytext.read(tmp_py), nb("saved from cm"))

    jupytext.write(nb("md edited"), tmp_md)
    model = await ensure_async(cm.get("notebook.ipynb"))
    compare_notebooks(model["content"], nb("md edited"))

    await ensure_async(cm.save(model=model, path="notebook.ipynb"))
    compare_notebooks(jupytext.read(tmp_ipynb), nb("md edited"))
    compare_notebooks(jupytext.read(tmp_md), nb("md edited"))
    compare_notebooks(jupytext.read(tmp_py), nb("md edited"))

    jupytext.write(nb("py edited"), tmp_py)

    # Loading the md file give the content of that file
    model = await ensure_async(cm.get("notebook.md"))
    compare_notebooks(model["content"], nb("md edited"))

    # Loading the ipynb files gives the content of the most recent text file
    model = await ensure_async(cm.get("notebook.ipynb"))
    compare_notebooks(model["content"], nb("py edited"))

    await ensure_async(cm.save(model=model, path="notebook.ipynb"))
    compare_notebooks(jupytext.read(tmp_ipynb), nb("py edited"))
    compare_notebooks(jupytext.read(tmp_md), nb("py edited"))
    compare_notebooks(jupytext.read(tmp_py), nb("py edited"))

    model_ipynb = await ensure_async(
        cm.get("notebook.ipynb", content=False, load_alternative_format=False)
    )
    model_md = await ensure_async(
        cm.get("notebook.md", content=False, load_alternative_format=False)
    )
    model_py = await ensure_async(
        cm.get("notebook.py", content=False, load_alternative_format=False)
    )

    # ipynb is the oldest one, then py, then md
    # so that we read cell inputs from the py file
    assert model_ipynb["last_modified"] <= model_py["last_modified"]
    assert model_py["last_modified"] <= model_md["last_modified"]


async def test_filter_jupytext_version_information_416(
    python_notebook, cm, tmpdir, cwd_tmpdir
):
    cm.root_dir = str(tmpdir)
    cm.notebook_metadata_filter = "-jupytext.text_representation.jupytext_version"

    # load notebook
    notebook = python_notebook
    notebook.metadata["jupytext_formats"] = "ipynb,py"
    model = notebook_model(notebook)

    # save to ipynb and py
    await ensure_async(cm.save(model=model, path="notebook.ipynb"))

    assert os.path.isfile("notebook.py")

    # read py file
    with open("notebook.py") as fp:
        text = fp.read()

    assert "---" in text
    assert "jupytext:" in text
    assert "kernelspec:" in text
    assert "jupytext_version:" not in text


@pytest.mark.requires_myst
async def test_new_untitled(tmpdir, cm):
    cm.root_dir = str(tmpdir)

    # untitled is "Untitled" only when the locale is English #636
    untitled, ext = (await ensure_async(cm.new_untitled(type="notebook")))[
        "path"
    ].split(".")
    assert untitled
    assert ext == "ipynb"

    # Jupytext related files
    assert (await ensure_async(cm.new_untitled(type="notebook", ext=".md")))[
        "path"
    ] == untitled + "1.md"
    assert (await ensure_async(cm.new_untitled(type="notebook", ext=".py")))[
        "path"
    ] == untitled + "2.py"
    assert (await ensure_async(cm.new_untitled(type="notebook", ext=".md:myst")))[
        "path"
    ] == untitled + "3.md"
    assert (await ensure_async(cm.new_untitled(type="notebook", ext=".py:percent")))[
        "path"
    ] == untitled + "4.py"
    assert (await ensure_async(cm.new_untitled(type="notebook", ext=".Rmd")))[
        "path"
    ] == untitled + "5.Rmd"

    # Test native formats that should not be changed by Jupytext and model should
    # not contain any Jupytext metadata neither file name should start with Uppercase
    for ext in [".py", ".md"]:
        model = await ensure_async(cm.new_untitled(type="file", ext=ext))
        assert model["content"] is None
        assert model["path"] == f"untitled{ext}"
    assert (await ensure_async(cm.new_untitled(type="directory")))[
        "path"
    ] == "Untitled Folder"


async def test_nested_prefix(tmpdir, cm):
    cm.root_dir = str(tmpdir)

    # save to ipynb and py
    nb = new_notebook(
        cells=[new_code_cell("1+1"), new_markdown_cell("Some text")],
        metadata={"jupytext": {"formats": "ipynb,nested/prefix//.py"}},
    )
    await ensure_async(cm.save(model=notebook_model(nb), path="notebook.ipynb"))

    assert tmpdir.join("nested").join("prefix").join("notebook.py").isfile()


async def test_timestamp_is_correct_after_reload_978(tmp_path, cm, python_notebook):
    """Here we reproduce the conditions in Issue #978 and make sure no
    warning is generated"""
    nb = python_notebook
    nb.metadata["jupytext"] = {"formats": "ipynb,py:percent"}

    cm.root_dir = str(tmp_path)

    ipynb_py_R_jl_file = tmp_path / "nb.ipynb"
    py_file = tmp_path / "nb.py"

    # 1. Save the paired notebook
    await ensure_async(cm.save(notebook_model(nb), path="nb.ipynb"))
    assert ipynb_py_R_jl_file.exists()
    assert py_file.exists()

    # and reload to get the original timestamp
    org_model = await ensure_async(cm.get("nb.ipynb"))

    # 2. Edit the py file
    time.sleep(0.5)
    text = py_file.read_text()
    text = (
        text
        + """

# %%
# A new cell
2 + 2
"""
    )

    py_file.write_text(text)

    # 3. Reload the paired notebook and make sure it has the modified content
    model = await ensure_async(cm.get("nb.ipynb"))
    nb = model["content"]
    assert "A new cell" in nb.cells[-1].source
    assert model["last_modified"] > org_model["last_modified"]


async def test_move_paired_notebook_to_subdir_1059(tmp_path, cm, python_notebook):
    (tmp_path / "jupytext.toml").write_text(
        'formats = "notebooks///ipynb,scripts///py:percent"\n'
    )
    cm.root_dir = str(tmp_path)

    # create paired notebook
    (tmp_path / "notebooks").mkdir()
    await ensure_async(
        cm.save(notebook_model(python_notebook), path="notebooks/my_notebook.ipynb")
    )
    assert (tmp_path / "notebooks" / "my_notebook.ipynb").exists()
    assert (tmp_path / "scripts" / "my_notebook.py").exists()

    # move notebook
    (tmp_path / "notebooks" / "subdir").mkdir()
    await ensure_async(
        cm.rename_file(
            "notebooks/my_notebook.ipynb", "notebooks/subdir/my_notebook.ipynb"
        )
    )
    assert (tmp_path / "notebooks" / "subdir" / "my_notebook.ipynb").exists()
    assert (tmp_path / "scripts" / "subdir" / "my_notebook.py").exists()

    assert not (tmp_path / "notebooks" / "my_notebook.ipynb").exists()
    assert not (tmp_path / "scripts" / "my_notebook.py").exists()

    # check notebook content
    model = await ensure_async(cm.get("scripts/subdir/my_notebook.py"))
    nb = model["content"]
    compare_notebooks(nb, python_notebook, fmt="py:percent")


async def test_hash_changes_if_paired_file_is_edited(tmp_path, cm, python_notebook):
    # 1. write py ipynb

    if "require_hash" not in inspect.signature(cm.get).parameters:
        pytest.skip(
            reason="This JupytextContentsManager does not have a 'require_hash' parameter in cm.get"
        )

    cm.formats = "ipynb,py:percent"
    cm.root_dir = str(tmp_path)

    # save ipynb
    nb = python_notebook
    nb_name = "notebook.ipynb"
    await ensure_async(cm.save(model=notebook_model(nb), path=nb_name))
    org_model = await ensure_async(cm.get(nb_name, require_hash=True))

    py_file = tmp_path / "notebook.py"

    text = py_file.read_text()
    assert "# %% [markdown]" in text.splitlines(), text

    # modify the timestamp of the paired file
    time.sleep(0.5)
    py_file.write_text(text)
    model = await ensure_async(cm.get(nb_name, require_hash=True))
    # not sure why the hash changes on Windows?
    assert (model["hash"] == org_model["hash"]) or (os.name == "nt")

    # modify the paired file
    py_file.write_text(text + "\n# %%\n1 + 1\n")

    new_model = await ensure_async(cm.get(nb_name, require_hash=True))
    assert new_model["hash"] != org_model["hash"]

    # the hash is for the pair (inputs first)
    model_from_py_file = await ensure_async(cm.get("notebook.py", require_hash=True))
    assert model_from_py_file["hash"] == new_model["hash"]



import pytest
from nbformat.v4.nbbase import new_markdown_cell, new_notebook

import jupytext
from jupytext.compare import compare


def test_markdown_cell_with_backslash_is_encoded_with_raw_string(
    nb=new_notebook(cells=[new_markdown_cell(r"A $\LaTeX$ expression")]),
    py=r'''# %% [markdown]
r"""
A $\LaTeX$ expression
"""
''',
):
    nb.metadata["jupytext"] = {
        "cell_markers": '"""',
        "notebook_metadata_filter": "-all",
    }
    py2 = jupytext.writes(nb, "py:percent")
    compare(py2, py)


@pytest.mark.parametrize("r", ["r", "R"])
@pytest.mark.parametrize("triple_quote", ['"""', "'''"])
@pytest.mark.parametrize("expr", ["$\\LaTeX$", "common"])
def test_raw_string_is_stable_over_round_trip(r, triple_quote, expr):
    py = f"""# %% [markdown]
{r}{triple_quote}
A {expr} expression
{triple_quote}
"""

    nb = jupytext.reads(py, "py:percent")

    (cell,) = nb.cells
    assert cell.cell_type == "markdown"
    assert cell.source == f"A {expr} expression"
    assert cell.metadata["cell_marker"] == f"{r}{triple_quote}"

    py2 = jupytext.writes(nb, "py:percent")
    compare(py2, py)



import contextlib
import itertools
import os.path
import re
import sys
import unittest.mock as mock
from pathlib import Path

import pytest
from jupyter_client.kernelspec import find_kernel_specs, get_kernel_spec
from nbformat.v4 import nbbase
from nbformat.v4.nbbase import (
    new_code_cell,
    new_markdown_cell,
    new_notebook,
    new_output,
)

import jupytext
from jupytext.cell_reader import rst2md
from jupytext.cli import system, tool_version
from jupytext.formats import formats_with_support_for_cell_metadata
from jupytext.myst import is_myst_available
from jupytext.pandoc import is_pandoc_available
from jupytext.quarto import is_quarto_available

# Pytest's tmpdir is in /tmp (at least for me), so this helps to avoid interferences between
# global configuration on HOME and the test collection
jupytext.config.JUPYTEXT_CEILING_DIRECTORIES = ["/tmp/"]

SAMPLE_NOTEBOOK_PATH = Path(__file__).parent / "data" / "notebooks" / "inputs"
ROOT_PATH = Path(__file__).parent.parent


@pytest.fixture(params=["sync", "async"])
def cm(request):
    if request.param == "sync":
        return jupytext.TextFileContentsManager()
    else:
        return jupytext.AsyncTextFileContentsManager()


@pytest.fixture
def no_jupytext_version_number():
    with mock.patch("jupytext.header.INSERT_AND_CHECK_VERSION_NUMBER", False):
        yield


@pytest.fixture
def cwd_tmpdir(tmpdir):
    # Run the whole test from inside tmpdir
    with tmpdir.as_cwd():
        yield tmpdir


@pytest.fixture()
def cwd_tmp_path(tmp_path):
    # Run the whole test from inside tmp_path
    if sys.version_info < (3, 11):
        with tmp_path.cwd():
            yield tmp_path
    else:
        # https://github.com/mwouts/jupytext/issues/1242
        with contextlib.chdir(tmp_path):
            yield tmp_path


@pytest.fixture
def jupytext_repo_root():
    """The local path of this repo, to use in .pre-commit-config.yaml in tests"""
    return str(Path(__file__).parent.parent.resolve())


@pytest.fixture
def jupytext_repo_rev(jupytext_repo_root):
    """The local revision of this repo, to use in .pre-commit-config.yaml in tests"""
    return system("git", "rev-parse", "HEAD", cwd=jupytext_repo_root).strip()


@pytest.fixture()
def python_notebook():
    return new_notebook(
        cells=[new_markdown_cell("A short notebook")],
        # We need a Python kernel here otherwise Jupytext is going to add
        # the information that this is a Python notebook when we sync the
        # .py and .ipynb files
        metadata={
            "kernelspec": {
                "display_name": "Python 3",
                "language": "python",
                "name": "python_kernel",
            },
            "language_info": {
                "codemirror_mode": {"name": "ipython", "version": 3},
                "file_extension": ".py",
                "mimetype": "text/x-python",
                "name": "python",
                "nbconvert_exporter": "python",
                "pygments_lexer": "ipython3",
                "version": "3.6.4",
            },
        },
    )


@pytest.fixture()
def notebook_with_outputs():
    return new_notebook(
        cells=[
            new_code_cell(
                "1+1",
                execution_count=1,
                outputs=[
                    new_output(
                        data={"text/plain": "2"},
                        execution_count=1,
                        output_type="execute_result",
                    )
                ],
            )
        ],
        metadata={
            "kernelspec": {
                "display_name": "Python 3",
                "language": "python",
                "name": "python_kernel",
            }
        },
    )


@pytest.fixture(params=list(formats_with_support_for_cell_metadata()))
def fmt_with_cell_metadata(request):
    return request.param


def list_notebooks(path="ipynb", skip=""):
    """All notebooks in the directory notebooks/path,
    or in the package itself"""
    if path == "ipynb":
        return (
            list_notebooks("ipynb_julia", skip=skip)
            + list_notebooks("ipynb_py", skip=skip)
            + list_notebooks("ipynb_R", skip=skip)
        )

    nb_path = SAMPLE_NOTEBOOK_PATH

    if path == "ipynb_all":
        return itertools.chain(
            *(
                list_notebooks(folder.name, skip=skip)
                for folder in nb_path.iterdir()
                if folder.name.startswith("ipynb_")
            )
        )

    if path == "all":
        return itertools.chain(
            *(list_notebooks(folder.name, skip=skip) for folder in nb_path.iterdir())
        )

    if path.startswith("."):
        nb_path = Path(__file__).parent / ".." / path
    else:
        nb_path = nb_path / path

    if skip:
        skip_re = re.compile(".*" + skip + ".*")
        return [
            str(nb_file)
            for nb_file in nb_path.iterdir()
            if nb_file.is_file() and not skip_re.match(nb_file.name)
        ]

    return [str(nb_file) for nb_file in nb_path.iterdir() if nb_file.is_file()]


def notebook_id_func(nb_file):
    nb_file = Path(nb_file)
    if SAMPLE_NOTEBOOK_PATH in nb_file.parents:
        return str(nb_file.relative_to(SAMPLE_NOTEBOOK_PATH))
    return str(nb_file.relative_to(ROOT_PATH))


@pytest.fixture(params=list_notebooks("ipynb_all"), ids=notebook_id_func)
def ipynb_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("all"), ids=notebook_id_func)
def any_nb_file(request):
    return request.param


@pytest.fixture
def ipynb_py_R_jl_files():
    return list_notebooks()


@pytest.fixture(params=list_notebooks(), ids=notebook_id_func)
def ipynb_py_R_jl_file(request):
    return request.param


@pytest.fixture
def ipynb_py_R_jl_ext(ipynb_py_R_jl_file):
    for language in "py", "R", "julia":
        if f"{os.path.sep}ipynb_{language}{os.path.sep}" in ipynb_py_R_jl_file:
            return ".jl" if language == "julia" else "." + language

    raise RuntimeError(f"language not found for {ipynb_py_R_jl_file}")


@pytest.fixture(
    params=list_notebooks("ipynb") + list_notebooks("Rmd"), ids=notebook_id_func
)
def ipynb_or_rmd_file(request):
    return request.param


@pytest.fixture(
    params=list_notebooks("ipynb_py") + list_notebooks("ipynb_R"), ids=notebook_id_func
)
def ipynb_py_R_file(request):
    return request.param


@pytest.fixture
def ipynb_py_files():
    return list_notebooks("ipynb_py")


@pytest.fixture(params=list_notebooks("ipynb_py"), ids=notebook_id_func)
def ipynb_py_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("ipynb_R"), ids=notebook_id_func)
def ipynb_R_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("ipynb_julia"), ids=notebook_id_func)
def ipynb_julia_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("ipynb_scheme"), ids=notebook_id_func)
def ipynb_scheme_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("ipynb_cpp"), ids=notebook_id_func)
def ipynb_cpp_file(request):
    return request.param


@pytest.fixture(
    params=list_notebooks("ipynb_all", skip="many hash"), ids=notebook_id_func
)
def ipynb_to_light(request):
    return request.param


@pytest.fixture(params=list_notebooks("ipynb_all"), ids=notebook_id_func)
def ipynb_to_myst(request):
    return request.param


@pytest.fixture(
    params=[
        py_file
        for py_file in list_notebooks("./src/jupytext")
        if py_file.endswith(".py") and "folding_markers" not in py_file
    ],
    ids=notebook_id_func,
)
def py_file(request):
    return request.param


@pytest.fixture(
    params=list_notebooks("julia")
    + list_notebooks("python")
    + list_notebooks("R")
    + list_notebooks("ps1"),
    ids=notebook_id_func,
)
def script_to_ipynb(request):
    return request.param


@pytest.fixture(params=list_notebooks("python"), ids=notebook_id_func)
def python_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("percent"), ids=notebook_id_func)
def percent_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("hydrogen"), ids=notebook_id_func)
def hydrogen_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("R"), ids=notebook_id_func)
def r_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("R_spin"), ids=notebook_id_func)
def r_spin_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("md"), ids=notebook_id_func)
def md_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("myst"), ids=notebook_id_func)
def myst_file(request):
    return request.param


@pytest.fixture(
    params=list_notebooks(
        "ipynb", skip="(functional|Notebook with|flavors|invalid|305)"
    ),
    ids=notebook_id_func,
)
def ipynb_to_pandoc(request):
    return request.param


@pytest.fixture(
    params=list_notebooks(
        "ipynb",
        skip="(functional|Notebook with|plotly_graphs|flavors|complex_metadata|"
        "update83|raw_cell|_66|nteract|LaTeX|invalid|305|text_outputs|ir_notebook|jupyter|with_R_magic)",
    ),
    ids=notebook_id_func,
)
def ipynb_to_quarto(request):
    return request.param


@pytest.fixture(
    params=list_notebooks("ipynb_py", skip="(raw|hash|frozen|magic|html|164|long)"),
    ids=notebook_id_func,
)
def ipynb_to_sphinx(request):
    return request.param


@pytest.fixture(params=list_notebooks("Rmd"), ids=notebook_id_func)
def rmd_file(request):
    return request.param


@pytest.fixture(params=list_notebooks("sphinx"), ids=notebook_id_func)
def sphinx_file(request):
    return request.param


def pytest_runtest_setup(item):
    for mark in item.iter_markers():
        for tool in [
            "jupytext",
            "black",
            "isort",
            "flake8",
            "autopep8",
            "jupyter nbconvert",
        ]:
            if mark.name == f"requires_{tool.replace(' ', '_')}":
                if not tool_version(tool):
                    pytest.skip(f"{tool} is not installed")
        if mark.name == "requires_sphinx_gallery":
            if not rst2md:
                pytest.skip("sphinx_gallery is not available")
        if mark.name == "requires_pandoc":
            # The mirror files changed slightly when Pandoc 2.11 was introduced
            # https://github.com/mwouts/jupytext/commit/c07d919702999056ce47f92b74f63a15c8361c5d
            # The mirror files changed again when Pandoc 2.16 was introduced
            # https://github.com/mwouts/jupytext/pull/919/commits/1fa1451ecdaa6ad8d803bcb6fb0c0cf09e5371bf
            if not is_pandoc_available(min_version="3.0"):
                pytest.skip("pandoc>=3.0 is not available")
        if mark.name == "requires_quarto":
            if not is_quarto_available(min_version="0.2.0"):
                pytest.skip("quarto>=0.2 is not available")
        if mark.name == "requires_no_pandoc":
            if is_pandoc_available():
                pytest.skip("Pandoc is installed")
        if mark.name == "requires_ir_kernel":
            if not any(
                get_kernel_spec(name).language == "R" for name in find_kernel_specs()
            ):
                pytest.skip("irkernel is not installed")
        if mark.name == "requires_user_kernel_python3":
            if "python_kernel" not in find_kernel_specs():
                pytest.skip(
                    "Please run 'python -m ipykernel install --name python_kernel --user'"
                )
        if mark.name == "requires_myst":
            if not is_myst_available():
                pytest.skip("myst_parser not found")
        if mark.name == "requires_no_myst":
            if is_myst_available():
                pytest.skip("myst is available")
        if mark.name == "skip_on_windows":
            if sys.platform.startswith("win"):
                pytest.skip("Issue 489")
        if mark.name == "pre_commit":
            if sys.platform.startswith("win"):
                pytest.skip(
                    "OSError: [WinError 193] %1 is not a valid Win32 application"
                )
            if not (Path(__file__).parent.parent / ".git").is_dir():
                pytest.skip("Jupytext folder is not a git repository #814")


def pytest_collection_modifyitems(config, items):
    for item in items:
        if (config.rootdir / "tests" / "external" / "pre_commit") in item.path.parents:
            item.add_marker(pytest.mark.pre_commit)


@pytest.fixture(autouse=True)
def cell_id():
    """To make sure that cell ids are distinct we use a global counter.
    This solves https://github.com/mwouts/jupytext/issues/747"""
    local_cell_count = 0

    def enumerate_cell_ids():
        nonlocal local_cell_count
        local_cell_count += 1
        return f"cell-{local_cell_count}"

    nbbase.random_cell_id = enumerate_cell_ids



import shutil
from io import StringIO
from unittest import mock

import pytest

from jupytext import read, reads
from jupytext.cli import jupytext
from jupytext.version import __version__


@pytest.mark.requires_user_kernel_python3
@pytest.mark.requires_nbconvert
@pytest.mark.skip_on_windows
def test_pipe_nbconvert_execute(tmpdir):
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("notebook.py"))

    with open(tmp_py, "w") as fp:
        fp.write(
            """1 + 2
"""
        )

    jupytext(
        args=[
            tmp_py,
            "--to",
            "ipynb",
            "--pipe-fmt",
            "ipynb",
            "--pipe",
            "jupyter nbconvert --stdin --stdout --to notebook --execute",
        ]
    )

    nb = read(tmp_ipynb)
    assert len(nb.cells) == 1
    assert nb.cells[0].outputs[0]["data"] == {"text/plain": "3"}


@pytest.mark.requires_user_kernel_python3
@pytest.mark.requires_nbconvert
@pytest.mark.skip_on_windows
def test_pipe_nbconvert_execute_sync(tmpdir):
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("notebook.py"))

    with open(tmp_py, "w") as fp:
        fp.write(
            """1 + 2
"""
        )

    jupytext(
        args=[
            tmp_py,
            "--set-formats",
            "py,ipynb",
            "--sync",
            "--pipe-fmt",
            "ipynb",
            "--pipe",
            "jupyter nbconvert --stdin --stdout --to notebook --execute",
        ]
    )

    nb = read(tmp_ipynb)
    assert len(nb.cells) == 1
    assert nb.cells[0].outputs[0]["data"] == {"text/plain": "3"}


@pytest.mark.requires_user_kernel_python3
@pytest.mark.requires_nbconvert
@pytest.mark.skip_on_windows
def test_execute(tmpdir, caplog, capsys):
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("notebook.py"))

    with open(tmp_py, "w") as fp:
        fp.write(
            """1 + 2
"""
        )

    jupytext(args=[tmp_py, "--to", "ipynb", "--execute"])

    nb = read(tmp_ipynb)
    assert len(nb.cells) == 1
    assert nb.cells[0].outputs[0]["data"] == {"text/plain": "3"}


@pytest.mark.requires_user_kernel_python3
@pytest.mark.requires_nbconvert
def test_execute_readme_ok(tmpdir):
    tmp_md = str(tmpdir.join("notebook.md"))

    with open(tmp_md, "w") as fp:
        fp.write(
            """
A readme with correct instructions

```python
1 + 2
```
"""
        )

    jupytext(args=[tmp_md, "--execute"])


@pytest.mark.requires_user_kernel_python3
@pytest.mark.requires_nbconvert
@pytest.mark.skip_on_windows
def test_execute_readme_not_ok(tmpdir):
    tmp_md = str(tmpdir.join("notebook.md"))

    with open(tmp_md, "w") as fp:
        fp.write(
            """
A readme with incorrect instructions (a is not defined)

```python
a + 1
```
"""
        )

    import nbconvert

    with pytest.raises(
        nbconvert.preprocessors.execute.CellExecutionError, match="is not defined"
    ):
        jupytext(args=[tmp_md, "--execute"])


@pytest.mark.requires_user_kernel_python3
@pytest.mark.requires_nbconvert
@pytest.mark.skip_on_windows
def test_execute_sync(tmpdir, caplog, capsys):
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_py = str(tmpdir.join("notebook.py"))

    with open(tmp_py, "w") as fp:
        fp.write(
            """1 + 2
"""
        )

    jupytext(args=[tmp_py, "--set-formats", "py,ipynb", "--sync", "--execute"])

    nb = read(tmp_ipynb)
    assert len(nb.cells) == 1
    assert nb.cells[0].outputs[0]["data"] == {"text/plain": "3"}


@pytest.mark.requires_nbconvert
@pytest.mark.requires_ir_kernel
@pytest.mark.skip_on_windows
def test_execute_r(tmpdir, caplog, capsys):  # pragma: no cover
    tmp_ipynb = str(tmpdir.join("notebook.ipynb"))
    tmp_md = str(tmpdir.join("notebook.md"))

    with open(tmp_md, "w") as fp:
        fp.write(
            """```r
1 + 2 + 3
```
"""
        )

    jupytext(args=[tmp_md, "--to", "ipynb", "--execute"])

    nb = read(tmp_ipynb)
    assert len(nb.cells) == 1
    assert nb.cells[0].outputs[0]["data"]["text/markdown"] == "6"


@pytest.mark.requires_user_kernel_python3
@pytest.mark.requires_nbconvert
@pytest.mark.skip_on_windows
def test_execute_in_subfolder(tmpdir, caplog, capsys):
    subfolder = tmpdir.mkdir("subfolder")

    tmp_csv = str(subfolder.join("inputs.csv"))
    tmp_py = str(subfolder.join("notebook.py"))
    tmp_ipynb = str(subfolder.join("notebook.ipynb"))

    with open(tmp_csv, "w") as fp:
        fp.write("1\n2\n")

    with open(tmp_py, "w") as fp:
        fp.write(
            """import ast

with open('inputs.csv') as fp:
    text = fp.read()

sum(ast.literal_eval(line) for line in text.splitlines())
"""
        )

    jupytext(args=[tmp_py, "--to", "ipynb", "--execute"])

    nb = read(tmp_ipynb)
    assert len(nb.cells) == 3
    assert nb.cells[2].outputs[0]["data"] == {"text/plain": "3"}

    tmp2_py = str(tmpdir.mkdir("another_folder").join("notebook.py"))
    tmp2_ipynb = str(tmpdir.join("another_folder", "notebook.ipynb"))

    shutil.copy(tmp_py, tmp2_py)

    # Executing without run-path fails
    import nbconvert

    with pytest.raises(
        nbconvert.preprocessors.execute.CellExecutionError,
        match="No such file or directory: 'inputs.csv'",
    ):
        jupytext(args=[tmp2_py, "--to", "ipynb", "--execute"])

    # Raise if folder does not exists
    with pytest.raises(ValueError, match="is not a valid path"):
        jupytext(args=[tmp2_py, "--to", "ipynb", "--run-path", "wrong_path"])

    # Execute in full path
    jupytext(args=[tmp2_py, "--to", "ipynb", "--run-path", str(subfolder)])
    nb = read(tmp2_ipynb)
    assert len(nb.cells) == 3
    assert nb.cells[2].outputs[0]["data"] == {"text/plain": "3"}

    # Execute in path relative to notebook dir
    jupytext(args=[tmp2_py, "--to", "ipynb", "--run-path", "../subfolder"])
    nb = read(tmp2_ipynb)
    assert len(nb.cells) == 3
    assert nb.cells[2].outputs[0]["data"] == {"text/plain": "3"}


@pytest.fixture()
def sample_md_notebook():
    """This is a sample md notebook with an outdated version of Jupytext
    and no kernel information, to test #908"""
    return """---
jupyter:
  jupytext:
    text_representation:
      extension: .md
      format_name: markdown
      format_version: '1.1'
      jupytext_version: 1.1.0
---

```python
1 + 1
```
"""


@pytest.mark.requires_user_kernel_python3
def test_execute_text_file_does_update_the_metadata(sample_md_notebook, tmp_path):
    md_file = tmp_path / "nb.md"
    md_file.write_text(sample_md_notebook)

    jupytext([str(md_file), "--execute"])

    new_md_text = md_file.read_text()
    assert __version__ in new_md_text
    assert "kernelspec" in new_md_text


@pytest.mark.requires_user_kernel_python3
def test_cat_execute_does_not_update_the_metadata(sample_md_notebook, tmp_path):
    md_file = tmp_path / "nb.md"
    md_file.write_text(sample_md_notebook)

    # read md notebook on stdin - this does the same as
    # cat notebook.md | jupytext --execute
    with open(md_file) as fp, mock.patch("sys.stdin", fp):
        jupytext(["--execute"])

    new_md_text = md_file.read_text()
    assert __version__ not in new_md_text
    assert "kernelspec" not in new_md_text


@pytest.mark.requires_user_kernel_python3
@pytest.mark.skip_on_windows
@pytest.mark.filterwarnings("ignore")
def test_utf8_out_331(capsys, caplog):
    py = "from IPython.core.display import HTML; HTML(u'\xd7')"

    with mock.patch("sys.stdin", StringIO(py)):
        jupytext(["--to", "ipynb", "--execute", "-"])

    out, err = capsys.readouterr()

    assert err == ""
    nb = reads(out, "ipynb")
    assert len(nb.cells) == 1
    print(nb.cells[0].outputs)
    assert nb.cells[0].outputs[0]["data"]["text/html"] == "\xd7"



"""Determine notebook or cell language"""
import re

# Jupyter magic commands that are also languages
_JUPYTER_LANGUAGES = [
    "R",
    "bash",
    "sh",
    "python",
    "python2",
    "python3",
    "coconut",
    "javascript",
    "js",
    "perl",
    "html",
    "latex",
    "markdown",
    "pypy",
    "ruby",
    "script",
    "svg",
    "matlab",
    "octave",
    "idl",
    "robotframework",
    "sas",
    "spark",
    "sql",
    "cython",
    "haskell",
    "tcl",
    "gnuplot",
    "wolfram language",
]

# Supported file extensions (and languages)
# Please add more languages here (and add a few tests) - see contributing.md
_SCRIPT_EXTENSIONS = {
    ".py": {"language": "python", "comment": "#"},
    ".coco": {"language": "coconut", "comment": "#"},
    ".R": {"language": "R", "comment": "#"},
    ".r": {"language": "R", "comment": "#"},
    ".jl": {"language": "julia", "comment": "#"},
    ".cpp": {"language": "c++", "comment": "//"},
    ".ss": {"language": "scheme", "comment": ";;"},
    ".clj": {"language": "clojure", "comment": ";;"},
    ".scm": {"language": "scheme", "comment": ";;"},
    ".sh": {"language": "bash", "comment": "#"},
    ".ps1": {"language": "powershell", "comment": "#"},
    ".q": {"language": "q", "comment": "/"},
    ".m": {"language": "matlab", "comment": "%"},
    # Unfortunately, Wolfram Mathematica also uses the .m extension which
    # conflicts with Matlab. To work around this problem we arbitrarily use a
    # made-up .wolfram extension.
    ".wolfram": {
        "language": "wolfram language",
        "comment": "(*",
        "comment_suffix": "*)",
    },
    ".pro": {"language": "idl", "comment": ";"},
    ".js": {"language": "javascript", "comment": "//"},
    ".ts": {"language": "typescript", "comment": "//"},
    ".scala": {"language": "scala", "comment": "//"},
    ".rs": {"language": "rust", "comment": "//"},
    ".robot": {"language": "robotframework", "comment": "#"},
    ".resource": {"language": "robotframework", "comment": "#"},
    ".cs": {"language": "csharp", "comment": "//"},
    ".fsx": {"language": "fsharp", "comment": "//"},
    ".fs": {"language": "fsharp", "comment": "//"},
    ".sos": {"language": "sos", "comment": "#"},
    ".java": {"language": "java", "comment": "//"},
    ".groovy": {"language": "groovy", "comment": "//"},
    ".sage": {"language": "sage", "comment": "#"},
    ".ml": {
        "language": "ocaml",
        "comment": "(*",
        "comment_suffix": "*)",
    },  # OCaml only has block comments
    ".hs": {"language": "haskell", "comment": "--"},
    ".tcl": {"language": "tcl", "comment": "#"},
    ".mac": {
        "language": "maxima",
        "comment": "/*",
        "comment_suffix": "*/",
    },  # Maxima only has block comments
    ".gp": {"language": "gnuplot", "comment": "#"},
    ".do": {"language": "stata", "comment": "//"},
    ".sas": {
        "language": "sas",
        "comment": "/*",
        "comment_suffix": "*/",
    },
    ".xsh": {"language": "xonsh", "comment": "#"},
    ".lgt": {"language": "logtalk", "comment": "%"},
    ".logtalk": {"language": "logtalk", "comment": "%"},
    ".lua": {"language": "lua", "comment": "--"},
    ".go": {"language": "go", "comment": "//"},
}

_COMMENT_CHARS = [
    _SCRIPT_EXTENSIONS[ext]["comment"]
    for ext in _SCRIPT_EXTENSIONS
    if _SCRIPT_EXTENSIONS[ext]["comment"] != "#"
]

_COMMENT = {
    _SCRIPT_EXTENSIONS[ext]["language"]: _SCRIPT_EXTENSIONS[ext]["comment"]
    for ext in _SCRIPT_EXTENSIONS
}
_JUPYTER_LANGUAGES = (
    set(_JUPYTER_LANGUAGES).union(_COMMENT.keys()).union(["c#", "f#", "cs", "fs"])
)
_JUPYTER_LANGUAGES_LOWER_AND_UPPER = _JUPYTER_LANGUAGES.union(
    {str.upper(lang) for lang in _JUPYTER_LANGUAGES}
)
_GO_DOUBLE_PERCENT_COMMAND = re.compile(r"^(%%\s*|%%\s+-.*)$")


def default_language_from_metadata_and_ext(metadata, ext, pop_main_language=False):
    """Return the default language given the notebook metadata, and a file extension"""
    default_from_ext = _SCRIPT_EXTENSIONS.get(ext, {}).get("language")

    main_language = metadata.get("jupytext", {}).get("main_language")
    default_language = (
        metadata.get("kernelspec", {}).get("language") or default_from_ext
    )
    language = main_language or default_language

    if (
        main_language is not None
        and main_language == default_language
        and pop_main_language
    ):
        metadata["jupytext"].pop("main_language")

    if language is None or language in ["R", "sas"]:
        return language

    if language.startswith("C++"):
        return "c++"

    return language.lower().replace("#", "sharp")


def usual_language_name(language):
    """Return the usual language name (one that may be found in _SCRIPT_EXTENSIONS above)"""
    language = language.lower()
    if language == "r":
        return "R"
    if language.startswith("c++"):
        return "c++"
    if language == "octave":
        return "matlab"
    if language in ["cs", "c#"]:
        return "csharp"
    if language in ["fs", "f#"]:
        return "fsharp"
    if language == "sas":
        return "SAS"
    return language


def same_language(kernel_language, language):
    """Are those the same language?"""
    return usual_language_name(kernel_language) == usual_language_name(language)


def set_main_and_cell_language(metadata, cells, ext, custom_cell_magics):
    """Set main language for the given collection of cells, and
    use magics for cells that use other languages"""
    main_language = default_language_from_metadata_and_ext(metadata, ext)

    if main_language is None:
        languages = {"python": 0.5}
        for cell in cells:
            if "language" in cell["metadata"]:
                language = usual_language_name(cell["metadata"]["language"])
                languages[language] = languages.get(language, 0.0) + 1

        main_language = max(languages, key=languages.get)

    # save main language when no kernel is set
    if "language" not in metadata.get("kernelspec", {}) and cells:
        metadata.setdefault("jupytext", {})["main_language"] = main_language

    # Remove 'language' meta data and add a magic if not main language
    for cell in cells:
        if "language" in cell["metadata"]:
            language = cell["metadata"]["language"]
            if language == main_language:
                cell["metadata"].pop("language")
                continue

            if usual_language_name(language) == main_language:
                continue

            if language in _JUPYTER_LANGUAGES or language in custom_cell_magics:
                cell["metadata"].pop("language")
                magic = "%%" if main_language != "csharp" else "#!"
                if "magic_args" in cell["metadata"]:
                    magic_args = cell["metadata"].pop("magic_args")
                    cell["source"] = (
                        f"{magic}{language} {magic_args}\n" + cell["source"]
                    )
                else:
                    cell["source"] = f"{magic}{language}\n" + cell["source"]


def cell_language(source, default_language, custom_cell_magics):
    """Return cell language and language options, if any"""
    if source:
        line = source[0]
        if default_language == "go" and _GO_DOUBLE_PERCENT_COMMAND.match(line):
            return None, None
        if default_language == "csharp":
            if line.startswith("#!"):
                lang = line[2:].strip()
                if lang in _JUPYTER_LANGUAGES:
                    source.pop(0)
                    return lang, ""
        elif line.startswith("%%"):
            magic = line[2:]
            if " " in magic:
                lang, magic_args = magic.split(" ", 1)
            else:
                lang = magic
                magic_args = ""

            if lang in _JUPYTER_LANGUAGES or lang in custom_cell_magics:
                source.pop(0)
                return lang, magic_args

    return None, None


def comment_lines(lines, prefix, suffix=""):
    """Return commented lines"""
    if not prefix:
        return lines
    if not suffix:
        return [prefix + " " + line if line else prefix for line in lines]
    return [
        prefix + " " + line + " " + suffix if line else prefix + " " + suffix
        for line in lines
    ]



import re
from pathlib import Path

from hatchling.metadata.plugin.interface import MetadataHookInterface


class AbsoluteLinksInReadme(MetadataHookInterface):
    """Hook that processes README.md to make relative links absolute."""

    def update(self, metadata):
        """Process README.md when metadata is being prepared."""
        readme_src_path = Path("README.md")
        readme_output_path = Path("build/README_with_absolute_links.md")
        base_url = "https://github.com/mwouts/jupytext/blob/main/"

        # Ensure the dist directory exists
        readme_output_path.parent.mkdir(exist_ok=True)

        self.convert_links(readme_src_path, readme_output_path, base_url)
        return metadata

    def convert_links(self, src_path, output_path, base_url):
        """Convert relative links in README.md to absolute links."""
        print(f"Processing {src_path} to generate {output_path} with absolute links")
        content = src_path.read_text(encoding="utf-8")

        # Find markdown links that don't start with http:// or https://
        pattern = re.compile(r"\[([^\]]+)\]\((?!http)([^)]+)\)")

        def replace_link(match):
            text, url = match.groups()
            if not url.startswith(("http://", "https://", "#")):
                url = base_url + url
            return f"[{text}]({url})"

        new_content = pattern.sub(replace_link, content)

        # Write the processed content to the output file
        output_path.parent.mkdir(exist_ok=True)
        output_path.write_text(new_content, encoding="utf-8")
        print(f"Generated {output_path} with absolute links")



import jupytext
from jupytext.compare import compare


# region Folding markers as cell boundaries
# region Sub-region with metadata {"key": "value"}
def test_mark_cell_with_vim_folding_markers(
    script="""# This is a markdown cell

# {{{ And this is a foldable code region with metadata {"key": "value"}
a = 1

b = 2

c = 3
# }}}
""",
):
    nb = jupytext.reads(script, "py")
    assert nb.metadata["jupytext"]["cell_markers"] == "{{{,}}}"
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "This is a markdown cell"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == "a = 1\n\nb = 2\n\nc = 3"
    assert nb.cells[1].metadata == {
        "title": "And this is a foldable code region with metadata",
        "key": "value",
    }

    script2 = jupytext.writes(nb, "py")
    compare(script2, script)


# endregion


def test_mark_cell_with_vscode_pycharm_folding_markers(
    script="""# This is a markdown cell

# region And this is a foldable code region with metadata {"key": "value"}
a = 1

b = 2

c = 3
# endregion
""",
):
    nb = jupytext.reads(script, "py")
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "This is a markdown cell"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == "a = 1\n\nb = 2\n\nc = 3"
    assert nb.cells[1].metadata == {
        "title": "And this is a foldable code region with metadata",
        "key": "value",
    }

    script2 = jupytext.writes(nb, "py")
    compare(script2, script)


def test_mark_cell_with_no_title_and_inner_region(
    script="""# This is a markdown cell

# region {"key": "value"}
a = 1

# region An inner region
b = 2
# endregion

def f(x):
    return x + 1


# endregion


d = 4
""",
):
    nb = jupytext.reads(script, "py")
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[0].source == "This is a markdown cell"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == '# region {"key": "value"}\na = 1'
    assert nb.cells[2].cell_type == "code"
    assert nb.cells[2].metadata["title"] == "An inner region"
    assert nb.cells[2].source == "b = 2"
    assert nb.cells[3].cell_type == "code"
    assert nb.cells[3].source == "def f(x):\n    return x + 1"
    assert nb.cells[4].cell_type == "code"
    assert nb.cells[4].source == "# endregion"
    assert nb.cells[5].cell_type == "code"
    assert nb.cells[5].source == "d = 4"
    assert len(nb.cells) == 6

    script2 = jupytext.writes(nb, "py")
    compare(script2, script)


# endregion


def test_adjacent_regions(
    script="""# region global
# region innermost
a = 1

b = 2
# endregion
# endregion
""",
):
    nb = jupytext.reads(script, "py")
    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "code"
    assert nb.cells[0].source == "# region global"
    assert nb.cells[1].cell_type == "code"
    assert nb.cells[1].source == "a = 1\n\nb = 2"
    assert nb.cells[2].cell_type == "code"
    assert nb.cells[2].source == "# endregion"

    script2 = jupytext.writes(nb, "py")
    compare(script2, script)


def test_indented_markers_are_ignored(
    script="""# region global
    # region indented
a = 1

b = 2
    # endregion
# endregion
""",
):
    nb = jupytext.reads(script, "py")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "code"

    script2 = jupytext.writes(nb, "py")
    compare(script2, script)



import jupytext
from jupytext.compare import compare


def test_read_simple_file(
    go="""// -*- coding: utf-8 -*-
// ---
// jupyter:
//   kernelspec:
//     display_name: Go (gonb)
//     language: go
//     name: gonb
// ---

// A notebook that use [GoNB](https://github.com/janpfeifer/gonb)

// the code below comes from [tutorial.ipynb](https://github.com/janpfeifer/gonb/blob/main/examples/tutorial.ipynb)

func main() {
    fmt.Printf("Hello World!")
}
""",
):
    nb = jupytext.reads(go, "go:light")
    assert len(nb.cells) == 3
    assert nb.cells[0].cell_type == "markdown"
    assert nb.cells[1].cell_type == "markdown"
    assert nb.cells[2].cell_type == "code"


def test_read_go_notebook_with_percent_percent_and_arguments(
    go="""// %% [markdown]
// # Square Function
//
// Defines a function $f(x) = x^2$

// %%
func Square(x float64) float64 {
        return x*x
}

// %% [markdown]
// # Examples
//
// ## Example A: $x = 3$

// %%
var x = flag.Float64("x", 0.0, "value of x to feed f(x)")

//gonb:%% -x=3
fmt.Printf("Square(%g)=%g\n", *x, Square(*x))

// %% [markdown]
// ## Example B: $x = 4$

// %%
//gonb:%% -x=4
fmt.Printf("Square(%g)=%g\n", *x, Square(*x))
""",
):
    nb = jupytext.reads(go, "go")
    for cell in nb.cells:
        assert "gonb" not in cell.source, cell.source

    go2 = jupytext.writes(nb, "go")
    compare(go2, go)


def test_read_go_notebook_with_magic_main(
    go="""// %%
package square

// %% [markdown]
// # Defining a $x^2$ function
// It returns x*x.

// %%
func Square(x float64) float64 {
  return x * x
}

// %% [markdown]
// # Examples
//
// ## Example A: $x = 3$

// %%
var x = flag.Float64("x", 0.0, "Value of x")

func example() {
  fmt.Printf("Square(%g)=%g\n", *x, Square(*x))
}
// %main example -x=3

// %% [markdown]
// ## Example B: $x = 4$

// %%
// %main example -x=4
""",
):
    nb = jupytext.reads(go, "go")
    for cell in nb.cells:
        assert "gonb" not in cell.source, cell.source

    go2 = jupytext.writes(nb, "go")
    compare(go2, go)


def test_commented_magic(
    go="""// %%
// This is a commented magic
// //gonb:%%

// %% [markdown]
// This is a commented magic in a markdown cell
// // //gonb:%%
""",
):
    nb = jupytext.reads(go, "go")
    assert len(nb.cells) == 2
    assert nb.cells[0].cell_type == "code"
    assert (
        nb.cells[0].source
        == """// This is a commented magic
// %%"""
    )
    assert nb.cells[1].cell_type == "markdown"
    assert (
        nb.cells[1].source
        == """This is a commented magic in a markdown cell
// %%"""
    )
    go2 = jupytext.writes(nb, "go")
    compare(go2, go)


def test_magic_commands_are_commented(
    go="""// %%
// !*rm -f go.work && go work init && go work use . ${HOME}/Projects/gopjrt
// %goworkfix
// %env LD_LIBRARY_PATH=/usr/local/lib
""",
):
    nb = jupytext.reads(go, "go:percent")
    assert len(nb.cells) == 1
    assert nb.cells[0].cell_type == "code"
    assert (
        nb.cells[0].source
        == """!*rm -f go.work && go work init && go work use . ${HOME}/Projects/gopjrt
%goworkfix
%env LD_LIBRARY_PATH=/usr/local/lib"""
    )

    go2 = jupytext.writes(nb, "go")
    compare(go2, go)



"""Determine how many blank lines should be inserted between two cells"""
from .stringparser import StringParser


def next_instruction_is_function_or_class(lines):
    """Is the first non-empty, non-commented line of the cell either a function or a class?"""
    parser = StringParser("python")
    for i, line in enumerate(lines):
        if parser.is_quoted():
            parser.read_line(line)
            continue
        parser.read_line(line)
        if not line.strip():  # empty line
            if i > 0 and not lines[i - 1].strip():
                return False
            continue
        if (
            line.startswith("def ")
            or line.startswith("async ")
            or line.startswith("class ")
        ):
            return True
        if line.startswith(("#", "@", " ", ")")):
            continue
        return False

    return False


def cell_ends_with_function_or_class(lines):
    """Does the last line of the cell belong to an indented code?"""
    non_quoted_lines = []
    parser = StringParser("python")
    for line in lines:
        if not parser.is_quoted():
            non_quoted_lines.append(line)
        parser.read_line(line)

    # find the first line, starting from the bottom, that is not indented
    lines = non_quoted_lines[::-1]
    for i, line in enumerate(lines):
        if not line.strip():
            # two blank lines? we won't need to insert more blank lines below this cell
            if i > 0 and not lines[i - 1].strip():
                return False
            continue
        if line.startswith(("#", " ", ")")):
            continue
        if (
            line.startswith("def ")
            or line.startswith("async ")
            or line.startswith("class ")
        ):
            return True
        return False

    return False


def cell_ends_with_code(lines):
    """Is the last line of the cell a line with code?"""
    if not lines:
        return False
    if not lines[-1].strip():
        return False
    if lines[-1].startswith("#"):
        return False
    return True


def cell_has_code(lines):
    """Is there any code in this cell?"""
    for i, line in enumerate(lines):
        stripped_line = line.strip()
        if stripped_line.startswith("#"):
            continue

        # Two consecutive blank lines?
        if not stripped_line:
            if i > 0 and not lines[i - 1].strip():
                return False
            continue

        return True

    return False


def pep8_lines_between_cells(prev_lines, next_lines, ext):
    """How many blank lines should be added between the two python paragraphs to make them pep8?"""
    if not next_lines:
        return 1
    if not prev_lines:
        return 0
    if ext != ".py":
        return 1
    if cell_ends_with_function_or_class(prev_lines):
        return 2 if cell_has_code(next_lines) else 1
    if cell_ends_with_code(prev_lines) and next_instruction_is_function_or_class(
        next_lines
    ):
        return 2
    return 1



import pytest

from jupytext import reads, writes
from jupytext.cli import pipe_notebook
from jupytext.compare import compare


@pytest.mark.requires_isort
def test_pipe_into_isort():
    text_org = """# %%
import numpy as np
np.array([1,2,3])

# %%
import pandas as pd
pd.Series([1,2,3])

# %%
# This is a comment on the second import
import pandas as pd
pd.Series([4,5,6])
"""

    text_target = """# %%
import numpy as np
# This is a comment on the second import
import pandas as pd

np.array([1,2,3])

# %%
pd.Series([1,2,3])

# %%
pd.Series([4,5,6])
"""

    nb_org = reads(text_org, fmt="py:percent")
    nb_pipe = pipe_notebook(
        nb_org, 'isort - --treat-comment-as-code "# %%" --float-to-top'
    )
    text_actual = writes(nb_pipe, "py:percent")
    compare(text_actual, text_target)



# ---
# jupyter:
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %%
%%html
<p><a href="https://github.com/mwouts/jupytext", style="color: rgb(0,0,255)">Jupytext</a> on GitHub</p>

# %%
%%latex
$\frac{\pi}{2}$

# %%
%load_ext rpy2.ipython

# %%
%%R
library(ggplot2)
ggplot(data=data.frame(x=c('A', 'B'), y=c(5, 2)), aes(x,weight=y)) + geom_bar()

# %%
%matplotlib inline
import pandas as pd
pd.Series({'A':5, 'B':2}).plot(figsize=(3,2), kind='bar')
